from __future__ import unicode_literals

import unittest

from fs.mode import validate_open_mode
from fs.mode import validate_openbin_mode
from fs import tools
from fs.opener import open_fs


class TestTools(unittest.TestCase):

    def test_remove_empty(self):
        fs = open_fs('temp://')
        fs.makedirs('foo/bar/baz/egg/')
        fs.create('foo/bar/test.txt')

        tools.remove_empty(fs, 'foo/bar/baz/egg')
        self.assertFalse(fs.isdir('foo/bar/baz'))
        self.assertTrue(fs.isdir('foo/bar'))
        fs.remove('foo/bar/test.txt')

        tools.remove_empty(fs, 'foo/bar')
        self.assertEqual(
            fs.listdir('/'),
            []
        )

    def test_validate_openbin_mode(self):
        with self.assertRaises(ValueError):
            validate_openbin_mode('X')
        with self.assertRaises(ValueError):
            validate_openbin_mode('')
        with self.assertRaises(ValueError):
            validate_openbin_mode('rX')
        with self.assertRaises(ValueError):
            validate_openbin_mode('rt')
        validate_openbin_mode('r')
        validate_openbin_mode('w')
        validate_openbin_mode('a')
        validate_openbin_mode('r+')
        validate_openbin_mode('w+')
        validate_openbin_mode('a+')

    def test_validate_open_mode(self):
        with self.assertRaises(ValueError):
            validate_open_mode('X')
        with self.assertRaises(ValueError):
            validate_open_mode('')
        with self.assertRaises(ValueError):
            validate_open_mode('rX')

        validate_open_mode('rt')
        validate_open_mode('r')
        validate_open_mode('rb')
        validate_open_mode('w')
        validate_open_mode('a')
        validate_open_mode('r+')
        validate_open_mode('w+')
        validate_open_mode('a+')




from __future__ import absolute_import
from __future__ import unicode_literals

import contextlib
import io
import os
import time

from collections import OrderedDict
from threading import RLock

import six

from . import errors
from .base import FS
from .enums import ResourceType
from .info import Info
from .path import iteratepath
from .path import normpath
from .path import split
from .mode import Mode


@six.python_2_unicode_compatible
class _MemoryFile(io.IOBase):

    def __init__(self, path, memory_fs, mode, dir_entry):
        super(_MemoryFile, self).__init__()
        self._path = path
        self._memory_fs = memory_fs
        self._mode = Mode(mode)
        self._dir_entry = dir_entry
        self._bytes_io = dir_entry.bytes_file

        self.accessed_time = time.time()
        self.modified_time = time.time()
        self.pos = 0

        if self._mode.truncate:
            with self._dir_entry.lock:
                self._bytes_io.seek(0)
                self._bytes_io.truncate()
        elif self._mode.appending:
            with self._dir_entry.lock:
                self._bytes_io.seek(0, os.SEEK_END)
                self.pos = self._bytes_io.tell()

    def __str__(self):
        _template = "<memoryfile '{path}' '{mode}'>"
        return _template.format(path=self._path, mode=self._mode)

    @contextlib.contextmanager
    def _seek_lock(self):
        with self._dir_entry.lock:
            self._bytes_io.seek(self.pos)
            yield
            self.pos = self._bytes_io.tell()

    def on_modify(self):
        """Called when file data is modified."""
        self._dir_entry.modified_time = self.modified_time = time.time()

    def on_access(self):
        """Called when file is accessed."""
        self._dir_entry.accessed_time = self.accessed_time = time.time()

    def flush(self):
        pass

    def __iter__(self):
        self._bytes_io.seek(self.pos)
        for line in self._bytes_io:
            yield line

    def next(self):
        with self._seek_lock():
            return next(self._bytes_io)

    __next__ = next

    def readline(self, *args, **kwargs):
        with self._seek_lock():
            self.on_access()
            return self._bytes_io.readline(*args, **kwargs)

    def close(self):
        if not self.closed:
            with self._dir_entry.lock:
                self._dir_entry.remove_open_file(self)
                super(_MemoryFile, self).close()

    def read(self, size=None):
        if not self._mode.reading:
            raise IOError('File not open for reading')
        if size is None:
            size = -1
        with self._seek_lock():
            self.on_access()
            return self._bytes_io.read(size)

    def readable(self):
        return self._mode.reading

    def readlines(self, hint=-1):
        with self._seek_lock():
            return self._bytes_io.readlines(hint)

    def seekable(self):
        return True

    def seek(self, *args, **kwargs):
        with self._seek_lock():
            self.on_access()
            return self._bytes_io.seek(*args, **kwargs)

    def tell(self):
        return self.pos

    def truncate(self, size=None):
        with self._seek_lock():
            self.on_modify()
            new_size = self._bytes_io.truncate(size)
            if size is not None and self._bytes_io.tell() < size:
                file_size = self._bytes_io.seek(0, os.SEEK_END)
                self._bytes_io.write(b'\0' * (size - file_size))
                self._bytes_io.seek(-size+file_size, os.SEEK_END)
            return size or new_size

    def writable(self):
        return self._mode.writing

    def write(self, data):
        if not self._mode.writing:
            raise IOError('File not open for writing')
        with self._seek_lock():
            self.on_modify()
            return self._bytes_io.write(data)

    def writelines(self, sequence):
        with self._seek_lock():
            self.on_modify()
            self._bytes_io.writelines(sequence)


class _DirEntry(object):

    def __init__(self, resource_type, name):
        self.resource_type = resource_type
        self.name = name
        self._dir = OrderedDict()
        self._open_files = []
        self._bytes_file = None
        self.lock = RLock()

        current_time = time.time()
        self.created_time = current_time
        self.accessed_time = current_time
        self.modified_time = current_time

        if not self.is_dir:
            self._bytes_file = io.BytesIO()

    @property
    def bytes_file(self):
        return self._bytes_file

    @property
    def is_dir(self):
        return self.resource_type == ResourceType.directory

    @property
    def size(self):
        with self.lock:
            if self.is_dir:
                return 0
            else:
                self._bytes_file.seek(0, os.SEEK_END)
                return self._bytes_file.tell()

    def get_entry(self, name, default=None):
        assert self.is_dir, 'must be a directory'
        return self._dir.get(name, default)

    def set_entry(self, name, dir_entry):
        self._dir[name] = dir_entry

    def remove_entry(self, name):
        del self._dir[name]

    def __contains__(self, name):
        return name in self._dir

    def __len__(self):
        return len(self._dir)

    def list(self):
        return list(self._dir.keys())

    def add_open_file(self, memory_file):
        self._open_files.append(memory_file)

    def remove_open_file(self, memory_file):
        self._open_files.remove(memory_file)


@six.python_2_unicode_compatible
class MemoryFS(FS):
    """
    A filesystem that stores all file and directory information in
    memory. This makes them very fast, but non-permanent.

    Memory filesystems are useful for caches, temporary data stores,
    unit testing, etc.

    Memory filesystems require no parameters to their constructor. The
    following is how you would create a ``MemoryFS`` instance::

        mem_fs = MemoryFS()

    """

    _meta = {
        'case_insensitive': False,
        'invalid_path_chars': '\0',
        'network': False,
        'read_only': False,
        'thread_safe': True,
        'unicode_paths': True,
        'virtual': False,
    }

    def __init__(self):
        """
        Create an in-memory filesystem.

        """
        self._meta = self._meta.copy()
        self.root = self._make_dir_entry(ResourceType.directory, '')
        super(MemoryFS, self).__init__()

    def __repr__(self):
        return "MemoryFS()"

    def __str__(self):
        return "<memfs>"

    def _make_dir_entry(self, *args, **kwargs):
        return _DirEntry(*args, **kwargs)

    def _get_dir_entry(self, dir_path):
        """Get a directory entry, or None if one doesn't exist."""
        with self._lock:
            dir_path = normpath(dir_path)
            current_entry = self.root
            for path_component in iteratepath(dir_path):
                if not current_entry.is_dir:
                    return None
                current_entry = current_entry.get_entry(path_component)
                if current_entry is None:
                    return None
            return current_entry

    def getinfo(self, path, namespaces=None):
        namespaces = namespaces or ()
        _path = self.validatepath(path)
        dir_entry = self._get_dir_entry(_path)
        if dir_entry is None:
            raise errors.ResourceNotFound(path)
        info = {
            'basic': {
                'name': dir_entry.name,
                'is_dir': dir_entry.is_dir
            }
        }
        if 'details' in namespaces:
            info['details'] = {
                "_write": ['accessed', 'modified'],
                "type": int(dir_entry.resource_type),
                "size": dir_entry.size,
                "accessed": dir_entry.accessed_time,
                "modified": dir_entry.modified_time,
                "created": dir_entry.created_time,
            }
        return Info(info)

    def listdir(self, path):
        self.check()
        _path = self.validatepath(path)
        with self._lock:
            dir_entry = self._get_dir_entry(_path)
            if dir_entry is None:
                raise errors.ResourceNotFound(path)
            if not dir_entry.is_dir:
                raise errors.DirectoryExpected(path)
            return dir_entry.list()

    def makedir(self, path, permissions=None, recreate=False):
        _path = self.validatepath(path)
        with self._lock:
            if _path == '/':
                if recreate:
                    return self.opendir(path)
                else:
                    raise errors.DirectoryExists(path)

            dir_path, dir_name = split(_path)

            parent_dir = self._get_dir_entry(dir_path)
            if parent_dir is None:
                raise errors.ResourceNotFound(path)

            dir_entry = parent_dir.get_entry(dir_name)
            if dir_entry is not None and not recreate:
                raise errors.DirectoryExists(path)

            if dir_entry is None:
                new_dir = self._make_dir_entry(
                    ResourceType.directory,
                    dir_name
                )
                parent_dir.set_entry(dir_name, new_dir)
            return self.opendir(path)

    def openbin(self, path, mode="r", buffering=-1, **options):
        _mode = Mode(mode)
        _mode.validate_bin()
        _path = self.validatepath(path)
        dir_path, file_name = split(_path)

        with self._lock:
            parent_dir_entry = self._get_dir_entry(dir_path)
            if parent_dir_entry is None or not parent_dir_entry.is_dir:
                raise errors.ResourceNotFound(path)

            if _mode.create:
                if file_name not in parent_dir_entry:
                    file_dir_entry = self._make_dir_entry(
                        ResourceType.file,
                        file_name
                    )
                    parent_dir_entry.set_entry(file_name, file_dir_entry)
                else:
                    file_dir_entry = self._get_dir_entry(_path)
                    if _mode.exclusive:
                        raise errors.FileExists(path)

                if file_dir_entry.is_dir:
                    raise errors.FileExpected(path)

                mem_file = _MemoryFile(
                    path=_path,
                    memory_fs=self,
                    mode=mode,
                    dir_entry=file_dir_entry
                )

                file_dir_entry.add_open_file(mem_file)
                return mem_file

            if file_name not in parent_dir_entry:
                raise errors.ResourceNotFound(path)

            file_dir_entry = parent_dir_entry.get_entry(file_name)
            if file_dir_entry.is_dir:
                raise errors.FileExpected(path)

            mem_file = _MemoryFile(
                path=_path,
                memory_fs=self,
                mode=mode,
                dir_entry=file_dir_entry
            )
            file_dir_entry.add_open_file(mem_file)
            return mem_file

    def remove(self, path):
        _path = self.validatepath(path)

        with self._lock:
            dir_path, file_name = split(_path)
            parent_dir_entry = self._get_dir_entry(dir_path)

            if parent_dir_entry is None or file_name not in parent_dir_entry:
                raise errors.ResourceNotFound(path)

            file_dir_entry = self._get_dir_entry(_path)
            if file_dir_entry.is_dir:
                raise errors.FileExpected(path)

            parent_dir_entry.remove_entry(file_name)

    def removedir(self, path):
        _path = self.validatepath(path)

        if _path == '/':
            raise errors.RemoveRootError()

        with self._lock:
            dir_path, file_name = split(_path)
            parent_dir_entry = self._get_dir_entry(dir_path)

            if parent_dir_entry is None or file_name not in parent_dir_entry:
                raise errors.ResourceNotFound(path)

            dir_dir_entry = self._get_dir_entry(_path)
            if not dir_dir_entry.is_dir:
                raise errors.DirectoryExpected(path)

            if len(dir_dir_entry):
                raise errors.DirectoryNotEmpty(path)

            parent_dir_entry.remove_entry(file_name)

    def setinfo(self, path, info):
        _path = self.validatepath(path)
        with self._lock:
            dir_path, file_name = split(_path)
            parent_dir_entry = self._get_dir_entry(dir_path)

            if parent_dir_entry is None or file_name not in parent_dir_entry:
                raise errors.ResourceNotFound(path)

            resource_entry = parent_dir_entry.get_entry(file_name)

            if 'details' in info:
                details = info['details']
                if 'accessed' in details:
                    resource_entry.accessed_time = details['accessed']
                if 'modified' in details:
                    resource_entry.modified_time = details['modified']



from __future__ import print_function
from __future__ import unicode_literals

import errno
from contextlib import contextmanager
import sys
import platform

from . import errors

from six import reraise


_WINDOWS_PLATFORM = platform.system() == 'Windows'


class _ConvertOSErrors(object):
    """Context manager to convert OSErrors in to FS Errors."""

    FILE_ERRORS = {
        64: errors.RemoteConnectionError,  # ENONET
        errno.EACCES: errors.PermissionDenied,
        errno.ENOENT: errors.ResourceNotFound,
        errno.EFAULT: errors.ResourceNotFound,
        errno.ESRCH: errors.ResourceNotFound,
        errno.ENOTEMPTY: errors.DirectoryNotEmpty,
        errno.EEXIST: errors.FileExists,
        183: errors.DirectoryExists,
        #errno.ENOTDIR: errors.DirectoryExpected,
        errno.ENOTDIR: errors.ResourceNotFound,
        errno.EISDIR: errors.FileExpected,
        errno.EINVAL: errors.FileExpected,
        errno.ENOSPC: errors.InsufficientStorage,
        errno.EPERM: errors.PermissionDenied,
        errno.ENETDOWN: errors.RemoteConnectionError,
        errno.ECONNRESET: errors.RemoteConnectionError,
        errno.ENAMETOOLONG: errors.PathError,
        errno.EOPNOTSUPP: errors.Unsupported,
        errno.ENOSYS: errors.Unsupported,
    }

    DIR_ERRORS = FILE_ERRORS.copy()
    DIR_ERRORS[errno.ENOTDIR] = errors.DirectoryExpected
    DIR_ERRORS[errno.EEXIST] = errors.DirectoryExists
    DIR_ERRORS[errno.EINVAL] = errors.DirectoryExpected

    if _WINDOWS_PLATFORM:  # pragma: no cover
        DIR_ERRORS[13] = errors.DirectoryExpected
        DIR_ERRORS[267] = errors.DirectoryExpected
        FILE_ERRORS[13] = errors.FileExpected

    def __init__(self, opname, path, directory=False):
        self._opname = opname
        self._path = path
        self._directory = directory

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        os_errors = (
            self.DIR_ERRORS
            if self._directory
            else self.FILE_ERRORS
        )
        if exc_type and isinstance(exc_value, EnvironmentError):
            _errno = exc_value.errno
            fserror = os_errors.get(_errno, errors.OperationFailed)
            if _errno == errno.EACCES and sys.platform == "win32":
                if getattr(exc_value, 'args', None) == 32:  # pragma: no cover
                    fserror = errors.ResourceLocked
            reraise(
                fserror,
                fserror(
                    self._path,
                    exc=exc_value
                ),
                traceback
            )

# Stops linter complaining about invalid class name
convert_os_errors = _ConvertOSErrors


@contextmanager
def unwrap_errors(path_replace):
    """
    A context manager to re-write the paths in resource exceptions to be
    in the same context as the wrapped filesystem.

    The only parameter may be the path from the parent, if only one path
    is to be unwrapped. Or it may be a dictionary that maps wrapped
    paths on to unwrapped paths.

    """
    try:
        yield
    except errors.ResourceError as e:
        if hasattr(e, 'path'):
            if isinstance(path_replace, dict):
                e.path = path_replace.get(e.path, e.path)
            else:
                e.path = path_replace
        reraise(type(e), e)



from __future__ import unicode_literals

import six

from . import errors
from .base import FS
from .copy import copy_file
from .info import Info
from .move import move_file
from .path import abspath, normpath
from .error_tools import unwrap_errors


@six.python_2_unicode_compatible
class WrapFS(FS):
    """"
    A proxy for a filesystem object.

    This class exposes an filesystem interface, where the data is
    stored on another filesystem(s), and is the basis for
    :class:`~fs.subfs.SubFS` and other *virtual* filesystems.

    """

    wrap_name = None

    def __init__(self, wrap_fs):
        self._wrap_fs = wrap_fs
        super(WrapFS, self).__init__()

    def __repr__(self):
        return "{}({!r})".format(
            self.__class__.__name__,
            self._wrap_fs
        )

    def __str__(self):
        wraps = []
        _fs = self
        while hasattr(_fs, '_wrap_fs'):
            wrap_name = getattr(_fs, 'wrap_name', None)
            if wrap_name is not None:
                wraps.append(wrap_name)
            _fs = _fs._wrap_fs
        if wraps:
            _str = "{}({})".format(_fs, ', '.join(wraps[::-1]))
        else:
            _str = "{}".format(_fs)
        return _str

    def delegate_path(self, path):
        """
        Encode a path for proxied filesystem.

        :param path: A path on the fileystem.
        :type path: str
        :returns: a tuple of <filesystem>, <new path>
        :rtype: tuple

        """
        return self._wrap_fs, path

    def delegate_fs(self):
        """
        Get the filesystem.

        This method should return a filesystem for methods not
        associated with a path, e.g. :meth:`~fs.base.FS.getmeta`.

        """
        return self._wrap_fs

    def appendbytes(self, path, data):
        self.check()
        _fs, _path = self.delegate_path(path)
        with unwrap_errors(path):
            return _fs.appendbytes(_path, data)

    def appendtext(self, path, text,
                   encoding='utf-8', errors=None, newline=''):
        self.check()
        _fs, _path = self.delegate_path(path)
        with unwrap_errors(path):
            return _fs.appendtext(_path,
                                  text,
                                  encoding=encoding,
                                  errors=errors,
                                  newline=newline)

    def getinfo(self, path, namespaces=None):
        self.check()
        _fs, _path = self.delegate_path(path)
        with unwrap_errors(path):
            raw_info = _fs.getinfo(_path, namespaces=namespaces).raw
        if abspath(normpath(path)) == '/':
            raw_info = raw_info.copy()
            raw_info['basic']['name'] = ''
        info = Info(raw_info)
        return info

    def listdir(self, path):
        self.check()
        _fs, _path = self.delegate_path(path)
        with unwrap_errors(path):
            dir_list = _fs.listdir(_path)
        return dir_list

    def lock(self):
        self.check()
        _fs = self.delegate_fs()
        return _fs.lock()

    def makedir(self, path, permissions=None, recreate=False):
        self.check()
        _fs, _path = self.delegate_path(path)
        with unwrap_errors(path):
            return _fs.makedir(
                _path,
                permissions=permissions,
                recreate=recreate
            )

    def move(self, src_path, dst_path, overwrite=False):
        # A custom move permits a potentially optimized code path
        src_fs, _src_path = self.delegate_path(src_path)
        dst_fs, _dst_path = self.delegate_path(dst_path)
        with unwrap_errors({_src_path: src_path, _dst_path: dst_path}):
            if not overwrite and dst_fs.exists(_dst_path):
                raise errors.DestinationExists(_dst_path)
            move_file(src_fs, _src_path, dst_fs, _dst_path)

    def openbin(self, path, mode='r', buffering=-1, **options):
        self.check()
        _fs, _path = self.delegate_path(path)
        with unwrap_errors(path):
            bin_file = _fs.openbin(
                _path,
                mode=mode,
                buffering=-1,
                **options
            )
        return bin_file

    def remove(self, path):
        self.check()
        _fs, _path = self.delegate_path(path)
        with unwrap_errors(path):
            _fs.remove(_path)

    def removedir(self, path):
        self.check()
        _path = abspath(normpath(path))
        if _path == '/':
            raise errors.RemoveRootError()
        _fs, _path = self.delegate_path(path)
        with unwrap_errors(path):
            _fs.removedir(_path)

    def scandir(self, path, namespaces=None, page=None):
        self.check()
        _fs, _path = self.delegate_path(path)
        with unwrap_errors(path):
            for info in _fs.scandir(_path,
                                    namespaces=namespaces,
                                    page=page):
                yield info

    def setinfo(self, path, info):
        self.check()
        _fs, _path = self.delegate_path(path)
        return _fs.setinfo(_path, info)

    def settimes(self, path, accessed=None, modified=None):
        self.check()
        _fs, _path = self.delegate_path(path)
        with unwrap_errors(path):
            _fs.settimes(_path, accessed=accessed, modified=modified)

    def touch(self, path):
        self.check()
        _fs, _path = self.delegate_path(path)
        with unwrap_errors(path):
            _fs.touch(_path)

    def copy(self, src_path, dst_path, overwrite=False):
        src_fs, _src_path = self.delegate_path(src_path)
        dst_fs, _dst_path = self.delegate_path(dst_path)
        with unwrap_errors({_src_path: src_path, _dst_path: dst_path}):
            if not overwrite and dst_fs.exists(_dst_path):
                raise errors.DestinationExists(_dst_path)
            copy_file(src_fs, _src_path, dst_fs, _dst_path)

    def create(self, path, wipe=False):
        self.check()
        _fs, _path = self.delegate_path(path)
        with unwrap_errors(path):
            _fs.create(_path, wipe=wipe)

    def desc(self, path):
        self.check()
        _fs, _path = self.delegate_path(path)
        with unwrap_errors(path):
            desc = _fs.desc(_path)
        return desc

    def exists(self, path):
        self.check()
        _fs, _path = self.delegate_path(path)
        with unwrap_errors(path):
            exists = _fs.exists(_path)
        return exists

    def filterdir(self,
                  path,
                  exclude_dirs=False,
                  exclude_files=False,
                  files=None,
                  dirs=None,
                  namespaces=None,
                  page=None):
        self.check()
        _fs, _path = self.delegate_path(path)
        iter_files = iter(_fs.filterdir(
            _path,
            exclude_dirs=exclude_dirs,
            exclude_files=exclude_files,
            files=files,
            dirs=dirs,
            namespaces=namespaces,
            page=page
        ))
        with unwrap_errors(path):
            for info in iter_files:
                yield info

    def getbytes(self, path):
        self.check()
        _fs, _path = self.delegate_path(path)
        with unwrap_errors(path):
            _bytes = _fs.getbytes(_path)
        return _bytes

    def gettext(self, path, encoding=None, errors=None, newline=''):
        self.check()
        _fs, _path = self.delegate_path(path)
        with unwrap_errors(path):
            _text = _fs.gettext(_path,
                                encoding=encoding,
                                errors=errors,
                                newline=newline)
        return _text

    def getmeta(self, namespace='standard'):
        self.check()
        meta = self.delegate_fs().getmeta(namespace=namespace)
        return meta

    def getsize(self, path):
        self.check()
        _fs, _path = self.delegate_path(path)
        with unwrap_errors(path):
            size = _fs.getsize(_path)
        return size

    def getsyspath(self, path):
        self.check()
        _fs, _path = self.delegate_path(path)
        with unwrap_errors(path):
            sys_path = _fs.getsyspath(_path)
        return sys_path

    def gettype(self, path):
        self.check()
        _fs, _path = self.delegate_path(path)
        with unwrap_errors(path):
            _type = _fs.gettype(_path)
        return _type

    def geturl(self, path, purpose='download'):
        self.check()
        _fs, _path = self.delegate_path(path)
        with unwrap_errors(path):
            return _fs.geturl(_path, purpose=purpose)

    def hassyspath(self, path):
        self.check()
        _fs, _path = self.delegate_path(path)
        with unwrap_errors(path):
            has_sys_path = _fs.hassyspath(_path)
        return has_sys_path

    def hasurl(self, path, purpose='download'):
        self.check()
        _fs, _path = self.delegate_path(path)
        with unwrap_errors(path):
            has_url = _fs.hasurl(_path, purpose=purpose)
        return has_url

    def isdir(self, path):
        self.check()
        _fs, _path = self.delegate_path(path)
        with unwrap_errors(path):
            _isdir = _fs.isdir(_path)
        return _isdir

    def isfile(self, path):
        self.check()
        _fs, _path = self.delegate_path(path)
        with unwrap_errors(path):
            _isfile = _fs.isfile(_path)
        return _isfile

    def islink(self, path):
        self.check()
        _fs, _path = self.delegate_path(path)
        with unwrap_errors(path):
            _islink = _fs.islink(_path)
        return _islink

    def makedirs(self, path, permissions=None, recreate=False):
        self.check()
        _fs, _path = self.delegate_path(path)
        return _fs.makedirs(
            _path,
            permissions=permissions,
            recreate=recreate
        )

    def open(self,
             path,
             mode='r',
             buffering=-1,
             encoding=None,
             errors=None,
             newline='',
             line_buffering=False,
             **options):
        self.check()
        _fs, _path = self.delegate_path(path)
        with unwrap_errors(path):
            open_file = _fs.open(
                _path,
                mode=mode,
                buffering=buffering,
                encoding=encoding,
                errors=errors,
                newline=newline,
                line_buffering=line_buffering,
                **options
            )
        return open_file

    def opendir(self, path, factory=None):
        from .subfs import SubFS
        factory = factory or SubFS
        if not self.getinfo(path).is_dir:
            raise errors.DirectoryExpected(
                path=path
            )
        with unwrap_errors(path):
            return factory(self, path)

    def setbytes(self, path, contents):
        self.check()
        _fs, _path = self.delegate_path(path)
        with unwrap_errors(path):
            _fs.setbytes(_path, contents)

    def setbinfile(self, path, file):
        self.check()
        _fs, _path = self.delegate_path(path)
        with unwrap_errors(path):
            _fs.setbinfile(_path, file)

    def setfile(self,
                path,
                file,
                encoding=None,
                errors=None,
                newline=''):
        self.check()
        _fs, _path = self.delegate_path(path)
        with unwrap_errors(path):
            _fs.setfile(
                _path,
                file,
                encoding=encoding,
                errors=errors,
                newline=newline
            )

    def validatepath(self, path):
        self.check()
        _fs, _path = self.delegate_path(path)
        with unwrap_errors(path):
            _fs.validatepath(_path)



# coding: utf-8
"""Defines the ZipOpener."""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from .base import Opener

class ZipOpener(Opener):
    protocols = ['zip']

    def open_fs(self, fs_url, parse_result, writeable, create, cwd):
        from ..zipfs import ZipFS
        zip_fs = ZipFS(
            parse_result.resource,
            write=create
        )
        return zip_fs



"""Test f2s.fnmatch."""

from __future__ import unicode_literals

import unittest

from fs import wildcard


class TestFNMatch(unittest.TestCase):
    """Test wildcard."""

    def test_wildcard(self):
        self.assertTrue(wildcard.match('*.py', 'file.py'))
        self.assertTrue(wildcard.match('????.py', '????.py'))
        self.assertTrue(wildcard.match('file.py', 'file.py'))
        self.assertTrue(wildcard.match('file.py[co]', 'file.pyc'))
        self.assertTrue(wildcard.match('file.py[co]', 'file.pyo'))
        self.assertTrue(wildcard.match('file.py[!c]', 'file.py0'))
        self.assertTrue(wildcard.match('file.py[^]', 'file.py^'))

        self.assertFalse(wildcard.match('*.jpg', 'file.py'))
        self.assertFalse(wildcard.match('toolong.py', '????.py'))
        self.assertFalse(wildcard.match('file.pyc', 'file.py'))
        self.assertFalse(wildcard.match('file.py[co]', 'file.pyz'))
        self.assertFalse(wildcard.match('file.py[!o]', 'file.pyo'))
        self.assertFalse(wildcard.match('file.py[]', 'file.py0'))

        self.assertTrue(wildcard.imatch('*.py', 'FILE.py'))
        self.assertTrue(wildcard.imatch('*.py', 'file.PY'))

    def test_match_any(self):
        self.assertTrue(wildcard.match_any([], 'foo.py'))
        self.assertTrue(wildcard.imatch_any([], 'foo.py'))
        self.assertTrue(wildcard.match_any(['*.py', '*.pyc'], 'foo.pyc'))
        self.assertTrue(wildcard.imatch_any(['*.py', '*.pyc'], 'FOO.pyc'))

    def test_get_matcher(self):
        matcher = wildcard.get_matcher([], True)
        self.assertTrue(matcher('foo.py'))
        matcher = wildcard.get_matcher(['*.py'], True)
        self.assertTrue(matcher('foo.py'))
        self.assertFalse(matcher('foo.PY'))
        matcher = wildcard.get_matcher(['*.py'], False)
        self.assertTrue(matcher('foo.py'))
        self.assertTrue(matcher('FOO.py'))



# coding: utf-8
"""
fs.opener.registry
===================

Defines the Registry, which maps protocols and FS URLs to their
respective Opener.

"""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import contextlib
import six
import pkg_resources

from .base import Opener
from .errors import UnsupportedProtocol, EntryPointError
from .parse import parse_fs_url


class Registry(object):
    """
    A registry for `Opener` instances.

    """


    def __init__(self, default_opener='osfs'):
        """
        Create a registry object.

        :param default_opener: The protocol to use, if one is not
            supplied. The default is to use 'osfs', so that the FS URL
            is treated as a system path if no protocol is given.

        """
        self.default_opener = default_opener
        self._protocols = None


    def __repr__(self):
        return "<fs-registry {!r}>".format(self.protocols)

    @property
    def protocols(self):
        """A list of supported protocols."""
        if self._protocols is None:
            self._protocols = [
                entry_point.name
                for entry_point in
                pkg_resources.iter_entry_points('fs.opener')
            ]
        return self._protocols

    def get_opener(self, protocol):
        """
        Get the opener class associated to a given protocol.

        :param str protocol: A filesystem protocol.
        :rtype: ``Opener``.
        :raises `~fs.opener.errors.UnsupportedProtocol`: If no opener
            could be found.
        :raises `EntryPointLoadingError`: If the returned entry
            point is not an ``Opener`` subclass or could not be loaded
            successfully.

        """
        protocol = protocol or self.default_opener
        entry_point = next(
            pkg_resources.iter_entry_points('fs.opener', protocol),
            None
        )

        if entry_point is None:
            raise UnsupportedProtocol(
                "protocol '{}' is not supported".format(protocol)
            )

        try:
            opener = entry_point.load()
        except Exception as exception:
            six.raise_from(
                EntryPointError(
                    'could not load entry point; {}'.format(exception)
                ),
                exception
            )
        else:
            if not issubclass(opener, Opener):
                raise EntryPointError(
                    'entry point did not return an opener'
                )

        try:
            opener_instance = opener()
        except Exception as exception:
            six.raise_from(
                EntryPointError(
                    'could not instantiate opener; {}'.format(exception)
                ),
                exception
            )

        return opener_instance

    def open(self,
             fs_url,
             writeable=True,
             create=False,
             cwd=".",
             default_protocol='osfs'):
        """
        Open a filesystem from a FS URL. Returns a tuple of a filesystem
        object and a path. If there is no path in the FS URL, the path
        value will be ``None``.

        :param str fs_url: A filesystem URL
        :param bool writeable: True if the filesystem must be writeable.
        :param bool create: True if the filesystem should be created if
            it does not exist.
        :param cwd: The current working directory.
        :type cwd: str or None
        :rtype: Tuple of ``(<filesystem>, <path from url>)``

        """

        if '://' not in fs_url:
            # URL may just be a path
            fs_url = "{}://{}".format(default_protocol, fs_url)

        parse_result = parse_fs_url(fs_url)
        protocol = parse_result.protocol
        open_path = parse_result.path

        opener = self.get_opener(protocol)

        open_fs = opener.open_fs(
            fs_url,
            parse_result,
            writeable,
            create,
            cwd
        )
        return open_fs, open_path

    def open_fs(self,
                fs_url,
                writeable=True,
                create=False,
                cwd=".",
                default_protocol='osfs'):
        """
        Open a filesystem object from a FS URL (ignoring the path
        component).

        :param str fs_url: A filesystem URL.
        :param bool writeable: True if the filesystem must be writeable.
        :param bool create: True if the filesystem should be created if
            it does not exist.
        :param str cwd: The current working directory (generally only
            relevant for OS filesystems).
        :param str default_protocol: The protocol to use if one is not
            supplied in the FS URL (defaults to ``"osfs"``).
        :returns: :class:`~fs.base.FS` object

        """
        from ..base import FS
        if isinstance(fs_url, FS):
            _fs = fs_url
        else:
            _fs, _path = self.open(
                fs_url,
                writeable=writeable,
                create=create,
                cwd=cwd,
                default_protocol=default_protocol
            )
        return _fs

    @contextlib.contextmanager
    def manage_fs(self, fs_url, create=False, writeable=True, cwd='.'):
        '''
        A context manager opens / closes a filesystem.

        :param fs_url: A FS instance or a FS URL.
        :type fs_url: str or FS
        :param bool create: If ``True``, then create the filesystem if
            it doesn't already exist.
        :param bool writeable: If ``True``, then the filesystem should
            be writeable.
        :param str cwd: The current working directory, if opening a
            :class:`~fs.osfs.OSFS`.

        Sometimes it is convenient to be able to pass either a FS object
        *or* an FS URL to a function. This context manager handles the
        required logic for that.

        Here's an example::

            def print_ls(list_fs):
                """List a directory."""
                with manage_fs(list_fs) as fs:
                    print(" ".join(fs.listdir()))

        This function may be used in two ways. You may either pass
        either a ``str``, as follows::

            print_list('zip://projects.zip')

        Or, an FS instance::

            from fs.osfs import OSFS
            projects_fs = OSFS('~/')
            print_list(projects_fs)

        '''
        from ..base import FS
        if isinstance(fs_url, FS):
            yield fs_url
        else:
            _fs = self.open_fs(
                fs_url,
                create=create,
                writeable=writeable,
                cwd=cwd
            )
            try:
                yield _fs
            except:
                raise
            finally:
                _fs.close()


registry = Registry()



from __future__ import unicode_literals

import unittest

from six import text_type

from fs import wrapfs
from fs.opener import open_fs


class WrappedFS(wrapfs.WrapFS):
    wrap_name = 'test'


class TestWrapFS(unittest.TestCase):

    def setUp(self):
        self.wrapped_fs = open_fs('mem://')
        self.fs = WrappedFS(self.wrapped_fs)

    def test_encode(self):
        self.assertEqual((self.wrapped_fs, 'foo'), self.fs.delegate_path('foo'))
        self.assertEqual((self.wrapped_fs, 'bar'), self.fs.delegate_path('bar'))
        self.assertIs(self.wrapped_fs, self.fs.delegate_fs())

    def test_repr(self):
        self.assertEqual(repr(self.fs), 'WrappedFS(MemoryFS())')

    def test_str(self):
        self.assertEqual(
            text_type(self.fs),
            '<memfs>(test)'
        )
        self.assertEqual(
            text_type(wrapfs.WrapFS(open_fs('mem://'))),
            '<memfs>'
        )



from __future__ import unicode_literals

from datetime import datetime
import time
import unittest

from fs import _ftp_parse as ftp_parse


class TestFTPParse(unittest.TestCase):

    def test_parse_time(self):
        self.assertEqual(
            ftp_parse._parse_time('JUL 05 1974'),
            142214400.0
        )

        year = time.localtime().tm_year
        self.assertEqual(
            ftp_parse._parse_time('JUL 05 02:00'),
            1499220000.0
        )

        self.assertEqual(
            ftp_parse._parse_time("notadate"),
            None
        )

    def test_parse(self):
        self.assertEqual(ftp_parse.parse(['']), [])

    def test_parse_line(self):
        self.assertIs(ftp_parse.parse_line('not a dir'), None)

    def test_decode_linux(self):
        directory = """\
lrwxrwxrwx    1 0        0              19 Jan 18  2006 debian -> ./pub/mirror/debian
drwxr-xr-x   10 0        0            4096 Aug 03 09:21 debian-archive
lrwxrwxrwx    1 0        0              27 Nov 30  2015 debian-backports -> pub/mirror/debian-backports
drwxr-xr-x   12 0        0            4096 Sep 29 13:13 pub
-rw-r--r--    1 0        0              26 Mar 04  2010 robots.txt
drwxr-xr-x   8 foo      bar          4096 Oct  4 09:05 test
"""

        expected = [{u'access': {u'group': u'0',
              u'permissions': [u'g_r',
                               u'g_w',
                               u'g_x',
                               u'o_r',
                               u'o_w',
                               u'o_x',
                               u'u_r',
                               u'u_w',
                               u'u_x'],
              u'user': u'0'},
  u'basic': {u'is_dir': True, u'name': u'debian'},
  u'details': {u'modified': 1137542400.0, u'size': 19, u'type': 1},
  u'ftp': {u'ls': u'lrwxrwxrwx    1 0        0              19 Jan 18  2006 debian -> ./pub/mirror/debian'}},
 {u'access': {u'group': u'0',
              u'permissions': [u'g_r',
                               u'g_x',
                               u'o_r',
                               u'o_x',
                               u'u_r',
                               u'u_w',
                               u'u_x'],
              u'user': u'0'},
  u'basic': {u'is_dir': True, u'name': u'debian-archive'},
  u'details': {u'modified': 1501752060.0, u'size': 4096, u'type': 1},
  u'ftp': {u'ls': u'drwxr-xr-x   10 0        0            4096 Aug 03 09:21 debian-archive'}},
 {u'access': {u'group': u'0',
              u'permissions': [u'g_r',
                               u'g_w',
                               u'g_x',
                               u'o_r',
                               u'o_w',
                               u'o_x',
                               u'u_r',
                               u'u_w',
                               u'u_x'],
              u'user': u'0'},
  u'basic': {u'is_dir': True, u'name': u'debian-backports'},
  u'details': {u'modified': 1448841600.0, u'size': 27, u'type': 1},
  u'ftp': {u'ls': u'lrwxrwxrwx    1 0        0              27 Nov 30  2015 debian-backports -> pub/mirror/debian-backports'}},
 {u'access': {u'group': u'0',
              u'permissions': [u'g_r',
                               u'g_x',
                               u'o_r',
                               u'o_x',
                               u'u_r',
                               u'u_w',
                               u'u_x'],
              u'user': u'0'},
  u'basic': {u'is_dir': True, u'name': u'pub'},
  u'details': {u'modified': 1506690780.0, u'size': 4096, u'type': 1},
  u'ftp': {u'ls': u'drwxr-xr-x   12 0        0            4096 Sep 29 13:13 pub'}},
 {u'access': {u'group': u'0',
              u'permissions': [u'g_r', u'o_r', u'u_r', u'u_w'],
              u'user': u'0'},
  u'basic': {u'is_dir': False, u'name': u'robots.txt'},
  u'details': {u'modified': 1267660800.0, u'size': 26, u'type': 2},
  u'ftp': {u'ls': u'-rw-r--r--    1 0        0              26 Mar 04  2010 robots.txt'}},
 {u'access': {u'group': u'bar',
              u'permissions': [u'g_r',
                               u'g_x',
                               u'o_r',
                               u'o_x',
                               u'u_r',
                               u'u_w',
                               u'u_x'],
              u'user': u'foo'},
  u'basic': {u'is_dir': True, u'name': u'test'},
  u'details': {u'modified': 1507107900.0, u'size': 4096, u'type': 1},
  u'ftp': {u'ls': u'drwxr-xr-x   8 foo      bar          4096 Oct  4 09:05 test'}}]

        parsed = ftp_parse.parse(directory.splitlines())
        self.assertEqual(parsed, expected)



from __future__ import unicode_literals

import os
import shutil
import tempfile

from fs import osfs
from fs.path import relpath
from .test_osfs import TestOSFS


class TestSubFS(TestOSFS):
    """Test OSFS implementation."""

    def setUp(self):
        self.temp_dir = tempfile.mkdtemp("fstest")
        self.parent_fs = osfs.OSFS(self.temp_dir)
        self.parent_fs.makedir('__subdir__')
        self.fs = self.parent_fs.opendir('__subdir__')

    def tearDown(self):
        shutil.rmtree(self.temp_dir)
        self.parent_fs.close()
        self.fs.close()

    def _get_real_path(self, path):
        _path = os.path.join(self.temp_dir, '__subdir__', relpath(path))
        return _path



from __future__ import unicode_literals

from six import text_type

from fs.opener import open_fs
from fs import walk
from fs import errors
from fs.test import UNICODE_TEXT


class ArchiveTestCases(object):

    def make_source_fs(self):
        return open_fs('temp://')

    def build_source(self, fs):
        fs.makedirs('foo/bar/baz')
        fs.makedir('tmp')
        fs.settext('top.txt', 'Hello, World')
        fs.settext('top2.txt', 'Hello, World')
        fs.settext('foo/bar/egg', 'foofoo')
        fs.makedir('unicode')
        fs.settext('unicode/text.txt', UNICODE_TEXT)

    def compress(self, fs):
        pass

    def load_archive(self):
        pass

    def remove_archive(self):
        pass

    def setUp(self):
        self.source_fs = source_fs = self.make_source_fs()
        self.build_source(source_fs)
        self.compress(source_fs)
        self.fs = self.load_archive()

    def tearDown(self):
        self.source_fs.close()
        self.fs.close()
        self.remove_archive()

    def test_repr(self):
        repr(self.fs)

    def test_str(self):
        self.assertIsInstance(text_type(self.fs), text_type)

    def test_readonly(self):
        with self.assertRaises(errors.ResourceReadOnly):
            self.fs.makedir('newdir')
        with self.assertRaises(errors.ResourceReadOnly):
            self.fs.remove('top.txt')
        with self.assertRaises(errors.ResourceReadOnly):
            self.fs.removedir('foo/bar/baz')
        with self.assertRaises(errors.ResourceReadOnly):
            self.fs.create('foo.txt')
        with self.assertRaises(errors.ResourceReadOnly):
            self.fs.setinfo('foo.txt', {})

    def test_getinfo(self):
        root = self.fs.getinfo('/')
        self.assertEqual(root.name, '')
        self.assertTrue(root.is_dir)

        top = self.fs.getinfo('top.txt', 'details')
        self.assertEqual(top.size, 12)
        self.assertFalse(top.is_dir)

    def test_listdir(self):
        self.assertEqual(
            sorted(self.source_fs.listdir('/')),
            sorted(self.fs.listdir('/'))
        )
        with self.assertRaises(errors.DirectoryExpected):
            self.fs.listdir('top.txt')
        with self.assertRaises(errors.ResourceNotFound):
            self.fs.listdir('nothere')

    def test_open(self):
        with self.fs.open('top.txt') as f:
            chars = []
            while True:
                c = f.read(2)
                if not c:
                    break
                chars.append(c)
            self.assertEqual(
                ''.join(chars),
                'Hello, World'
            )
        with self.assertRaises(errors.ResourceNotFound):
            with self.fs.open('nothere.txt') as f:
                pass
        with self.assertRaises(errors.FileExpected):
            with self.fs.open('foo') as f:
                pass

    def test_gets(self):
        self.assertEqual(
            self.fs.gettext('top.txt'),
            'Hello, World'
        )
        self.assertEqual(
            self.fs.gettext('foo/bar/egg'),
            'foofoo'
        )
        self.assertEqual(
            self.fs.getbytes('top.txt'),
            b'Hello, World'
        )
        self.assertEqual(
            self.fs.getbytes('foo/bar/egg'),
            b'foofoo'
        )
        with self.assertRaises(errors.ResourceNotFound):
            self.fs.getbytes('what.txt')

    def test_walk_files(self):
        source_files = sorted(walk.walk_files(self.source_fs))
        archive_files = sorted(walk.walk_files(self.fs))

        self.assertEqual(
            source_files,
            archive_files
        )

    def test_implied_dir(self):
        self.fs.getinfo('foo/bar')
        self.fs.getinfo('foo')




"""Time related tools."""

from __future__ import print_function
from __future__ import unicode_literals

from calendar import timegm
from datetime import datetime
from pytz import UTC, timezone


utcfromtimestamp = datetime.utcfromtimestamp
utclocalize = UTC.localize
GMT = timezone('GMT')


def datetime_to_epoch(d):
    """Convert datetime to epoch."""
    return timegm(d.utctimetuple())


def epoch_to_datetime(t):
    """Convert epoch time to a UTC datetime."""
    return utclocalize(utcfromtimestamp(t)) if t is not None else None



"""

Functions for copying resources *between* filesystem.

"""

from __future__ import print_function
from __future__ import unicode_literals

from .walk import Walker
from .opener import manage_fs
from .path import abspath
from .path import combine
from .path import frombase
from .path import normpath
from .errors import FSError


def copy_fs(src_fs, dst_fs, walker=None, on_copy=None):
    """
    Copy the contents of one filesystem to another.

    :param src_fs: Source filesystem.
    :type src_fs: FS URL or instance
    :param dst_fs: Destination filesystem.
    :type dst_fs: FS URL or instance
    :param walker: A walker object that will be used to scan for files
        in ``src_fs``. Set this if you only want to consider a sub-set
        of the resources in ``src_fs``.
    :type walker: :class:`~fs.walk.Walker`
    :param on_copy: A function callback called after a single file copy
        is executed.
    :type on_copy: Function, with signature ``(src_fs, src_path, dst_fs,
        dst_path)``.
    """
    return copy_dir(src_fs, '/', dst_fs, '/',
                    walker=walker, on_copy=on_copy)


def copy_fs_if_newer(src_fs, dst_fs, walker=None, on_copy=None):
    """
    Copy the contents of one filesystem to another. If both source and
    destination files exist, the copy is executed only if the source
    file is newer than the destination file. In case modification times
    of source or destination files are not available, copy file is
    always executed.

    :param src_fs: Source filesystem.
    :type src_fs: FS URL or instance
    :param dst_fs: Destination filesystem.
    :type dst_fs: FS URL or instance
    :param walker: A walker object that will be used to scan for files
        in ``src_fs``. Set this if you only want to consider a sub-set
        of the resources in ``src_fs``.
    :type walker: :class:`~fs.walk.Walker`
    :param on_copy: A function callback called after a single file copy
        is executed.
    :type on_copy: Function, with signature ``(src_fs, src_path, dst_fs,
        dst_path)``.
    """
    return copy_dir_if_newer(src_fs, '/', dst_fs, '/',
                             walker=walker, on_copy=on_copy)


def _source_is_newer(src_fs, src_path, dst_fs, dst_path):
    """
    Determine if source file is newer than destination file.

    :param src_fs: Source filesystem.
    :type src_fs: FS URL or instance
    :param src_path: Path to a file on ``src_fs``.
    :type src_path: str
    :param dst_fs: Destination filesystem.
    :type dst_fs: FS URL or instance
    :param dst_path: Path to a file on ``dst_fs``.
    :type dst_path: str
    :returns: True if source file is newer than destination file or
        file modification time cannot be determined. False otherwise.
    """
    try:
        if dst_fs.exists(dst_path):
            namespace = ('details', 'modified')
            src_modified = src_fs.getinfo(src_path, namespace).modified
            if src_modified is not None:
                dst_modified = dst_fs.getinfo(dst_path, namespace).modified
                return dst_modified is None or src_modified > dst_modified
        return True
    except FSError:  # pragma: nocover
        #todo: should log something here
        return True


def copy_file(src_fs, src_path, dst_fs, dst_path):
    """
    Copy a file from one filesystem to another. If the destination
    exists, and is a file, it will be first truncated.

    :param src_fs: Source filesystem.
    :type src_fs: FS URL or instance
    :param src_path: Path to a file on ``src_fs``.
    :type src_path: str
    :param dst_fs: Destination filesystem.
    :type dst_fs: FS URL or instance
    :param dst_path: Path to a file on ``dst_fs``.
    :type dst_path: str

    """
    with manage_fs(src_fs, writeable=False) as src_fs:
        with manage_fs(dst_fs, create=True) as dst_fs:
            if src_fs is dst_fs:
                # Same filesystem, so we can do a potentially optimized
                # copy
                src_fs.copy(src_path, dst_path, overwrite=True)
            else:
                # Standard copy
                with src_fs.lock(), dst_fs.lock():
                    with src_fs.open(src_path, 'rb') as read_file:
                        # There may be an optimized copy available on
                        # dst_fs
                        dst_fs.setbinfile(dst_path, read_file)


def copy_file_if_newer(src_fs, src_path, dst_fs, dst_path):
    """
    Copy a file from one filesystem to another. If the destination
    exists, and is a file, it will be first truncated. If both source
    and destination files exist, the copy is executed only if the source
    file is newer than the destination file. In case modification times
    of source or destination files are not available, copy is always
    executed.

    :param src_fs: Source filesystem.
    :type src_fs: FS URL or instance
    :param src_path: Path to a file on ``src_fs``.
    :type src_path: str
    :param dst_fs: Destination filesystem.
    :type dst_fs: FS URL or instance
    :param dst_path: Path to a file on ``dst_fs``.
    :type dst_path: str
    :returns: True if the file copy was executed, False otherwise.

    """
    with manage_fs(src_fs, writeable=False) as src_fs:
        with manage_fs(dst_fs, create=True) as dst_fs:
            if src_fs is dst_fs:
                # Same filesystem, so we can do a potentially optimized
                # copy
                if _source_is_newer(src_fs, src_path, dst_fs, dst_path):
                    src_fs.copy(src_path, dst_path, overwrite=True)
                    return True
                else:
                    return False
            else:
                # Standard copy
                with src_fs.lock(), dst_fs.lock():
                    if _source_is_newer(src_fs, src_path,
                                        dst_fs, dst_path):
                        with src_fs.open(src_path, 'rb') as read_file:
                            # There may be an optimized copy available
                            # on dst_fs
                            dst_fs.setbinfile(dst_path, read_file)
                            return True
                    else:
                        return False



def copy_structure(src_fs, dst_fs, walker=None):
    """
    Copy directories (but not files) from ``src_fs`` to ``dst_fs``.

    :param src_fs: Source filesystem.
    :type src_fs: FS URL or instance
    :param dst_fs: Destination filesystem.
    :type dst_fs: FS URL or instance
    :param walker: A walker object that will be used to scan for files
        in ``src_fs``. Set this if you only want to consider a sub-set
        of the resources in ``src_fs``.
    :type walker: :class:`~fs.walk.Walker`

    """
    walker = walker or Walker()
    with manage_fs(src_fs, writeable=False) as src_fs:
        with manage_fs(dst_fs, create=True) as dst_fs:
            with src_fs.lock(), dst_fs.lock():
                for dir_path in walker.dirs(src_fs):
                    dst_fs.makedir(dir_path, recreate=True)


def copy_dir(src_fs, src_path, dst_fs, dst_path,
             walker=None, on_copy=None):
    """
    Copy a directory from one filesystem to another.

    :param src_fs: Source filesystem.
    :type src_fs: FS URL or instance
    :param src_path: A path to a directory on ``src_fs``.
    :type src_path: str
    :param dst_fs: Destination filesystem.
    :type dst_fs: FS URL or instance
    :param str dst_path: A path to a directory on ``dst_fs``.
    :param walker: A walker object that will be used to scan for files
        in ``src_fs``. Set this if you only want to consider a sub-set
        of the resources in ``src_fs``.
    :type walker: :class:`~fs.walk.Walker`
    :param on_copy: A function callback called after a single file copy
        is executed.
    :type on_copy: Function, with signature ``(src_fs, src_path, dst_fs,
        dst_path)``.

    """
    on_copy = on_copy or (lambda *args: None)
    walker = walker or Walker()
    _src_path = abspath(normpath(src_path))
    _dst_path = abspath(normpath(dst_path))
    with manage_fs(src_fs, writeable=False) as src_fs:
        with manage_fs(dst_fs, create=True) as dst_fs:
            with src_fs.lock(), dst_fs.lock():
                dst_fs.makedir(_dst_path, recreate=True)
                for dir_path, dirs, files in walker.walk(src_fs, _src_path):
                    copy_path = combine(
                        _dst_path,
                        frombase(_src_path, dir_path)
                    )
                    for info in dirs:
                        dst_fs.makedir(
                            info.make_path(copy_path),
                            recreate=True
                        )
                    for info in files:
                        src_path = info.make_path(dir_path)
                        dst_path = info.make_path(copy_path)
                        copy_file(
                            src_fs,
                            src_path,
                            dst_fs,
                            dst_path
                        )
                        on_copy(src_fs, src_path, dst_fs, dst_path)


def copy_dir_if_newer(src_fs, src_path, dst_fs, dst_path,
                      walker=None, on_copy=None):
    """
    Copy a directory from one filesystem to another. If both source and
    destination files exist, the copy is executed only if the source
    file is newer than the destination file. In case modification times
    of source or destination files are not available, copy is always
    executed.

    :param src_fs: Source filesystem.
    :type src_fs: FS URL or instance
    :param src_path: A path to a directory on ``src_fs``.
    :type src_path: str
    :param dst_fs: Destination filesystem.
    :type dst_fs: FS URL or instance
    :param str dst_path: A path to a directory on ``dst_fs``.
    :param walker: A walker object that will be used to scan for files
        in ``src_fs``. Set this if you only want to consider a sub-set
        of the resources in ``src_fs``.
    :type walker: :class:`~fs.walk.Walker`
    :param on_copy: A function callback called after a single file copy
        is executed.
    :type on_copy: Function, with signature ``(src_fs, src_path, dst_fs,
        dst_path)``.

    """
    on_copy = on_copy or (lambda *args: None)
    walker = walker or Walker()
    _src_path = abspath(normpath(src_path))
    _dst_path = abspath(normpath(dst_path))
    with manage_fs(src_fs, writeable=False) as src_fs:
        with manage_fs(dst_fs, create=True) as dst_fs:
            with src_fs.lock(), dst_fs.lock():
                dst_fs.makedir(_dst_path, recreate=True)
                namespace = ('details', 'modified')
                dst_state = {
                    path: info
                    for path, info in walker.info(dst_fs, _dst_path, namespace)
                    if info.is_file
                }
                src_state = [
                    (path, info)
                    for path, info in walker.info(src_fs, _src_path, namespace)
                ]
                for dir_path, copy_info in src_state:
                    copy_path = combine(
                        _dst_path,
                        frombase(_src_path, dir_path)
                    )
                    if copy_info.is_dir:
                        dst_fs.makedir(copy_path, recreate=True)
                    elif copy_info.is_file:
                        # dst file is present, try to figure out if copy
                        # is necessary
                        src_modified = copy_info.modified
                        do_copy = (
                            dir_path not in dst_state or
                            src_modified is None or
                            dst_state[dir_path].modified is None or
                            src_modified > dst_state[dir_path].modified
                        )
                        if do_copy:
                            copy_file(src_fs, dir_path, dst_fs, copy_path)
                            on_copy(src_fs, dir_path, dst_fs, copy_path)




"""
Tools for managing mode strings (as used in :meth:`~fs.base.FS.open` and
:meth:`~fs.base.FS.openbin`).

"""

from __future__ import print_function
from __future__ import unicode_literals

import six


# https://docs.python.org/3/library/functions.html#open
@six.python_2_unicode_compatible
class Mode(object):
    """
    A mode object provides properties that can be used to interrogate
    the `mode
    strings <https://docs.python.org/3/library/functions.html#open>`_ used
    when opening files.

    :param str mode: A *mode* string, as used by ``io.open``.
    :raises ValueError: If the mode string is invalid.

    Here's an example of typical use::

        >>> mode = Mode('rb')
        >>> mode.reading
        True
        >>> mode.writing
        False
        >>> mode.binary
        True
        >>> mode.text
        False


    """

    def __init__(self, mode):
        self._mode = mode
        self.validate()

    def __repr__(self):
        return "Mode({!r})".format(self._mode)

    def __str__(self):
        return self._mode

    def __contains__(self, character):
        """Check if a mode contains a given character."""
        return character in self._mode

    def to_platform(self):
        """
        Get a mode string for the current platform.

        Currently, this just removes the 'x' on PY2 because PY2 doesn't
        support exclusive mode.

        """
        return self._mode.replace('x', 'w') if six.PY2 else self._mode

    def to_platform_bin(self):
        """
        Get a *binary* mode string for the current platform.

        Currently, this just removes the 'x' on PY2 because PY2 doesn't
        support exclusive mode.

        """
        _mode = self.to_platform().replace('t', '')
        return _mode if 'b' in _mode else _mode + 'b'

    def validate(self, _valid_chars=frozenset('rwxtab+')):
        """
        Validate the mode string.

        :raises ValueError: if the mode contains invalid chars.

        """

        mode = self._mode
        if not mode:
            raise ValueError('mode must not be empty')
        if not _valid_chars.issuperset(mode):
            raise ValueError(
                "mode '{}' contains invalid characters".format(mode)
            )
        if mode[0] not in 'rwxa':
            raise ValueError(
                "mode must start with 'r', 'w', 'x', or 'a'"
            )
        if 't' in mode and 'b' in mode:
            raise ValueError(
                "mode can't be binary ('b') and text ('t')"
            )

    def validate_bin(self):
        """
        Validate a mode for opening a binary file.

        :raises ValueError: if the mode contains invalid chars.

        """
        self.validate()
        if 't' in self:
            raise ValueError('mode must be binary')

    @property
    def create(self):
        """Check if the mode would create a file."""
        return 'a' in self or 'w' in self or 'x' in self

    @property
    def reading(self):
        """Check if the mode permits reading."""
        return 'r' in self or '+' in self

    @property
    def writing(self):
        """Check if a mode permits writing."""
        return 'w' in self or 'a' in self or '+' in self or 'x' in self

    @property
    def appending(self):
        """Check if a mode permits appending."""
        return 'a' in self

    @property
    def updating(self):
        """Check if a mode permits updating (reading and writing)."""
        return '+' in self

    @property
    def truncate(self):
        """Check if a mode would truncate an existing file."""
        return 'w' in self or 'x' in self

    @property
    def exclusive(self):
        """Check if the mode require exclusive creation."""
        return 'x' in self

    @property
    def binary(self):
        """Check if a mode specifies binary."""
        return 'b' in self

    @property
    def text(self):
        """Check if a mode specifies text."""
        return 't' in self or 'b' not in self


def check_readable(mode):
    """
    Check a mode string allows reading.

    :param str mode: A mode string, e.g. ``"rt"``
    :rtype: bool

    """
    return Mode(mode).reading


def check_writable(mode):
    """
    Check a mode string allows writing.

    :param str mode: A mode string, e.g. ``"wt"``
    :rtype: bool

    """
    return Mode(mode).writing


def validate_open_mode(mode):
    """
    Check ``mode`` parameter of :meth:`~fs.base.FS.open` is valid.

    :param str mode: Mode parameter.
    :raises: `ValueError` if mode is not valid.

    """
    Mode(mode)


def validate_openbin_mode(mode, _valid_chars=frozenset('rwxab+')):
    """
    Check ``mode`` parameter of :meth:`~fs.base.FS.openbin` is valid.

    :param mode: Mode parameter.
    :type mode: str
    :raises: `ValueError` if mode is not valid.

    """
    if 't' in mode:
        raise ValueError('text mode not valid in openbin')
    if not mode:
        raise ValueError('mode must not be empty')
    if mode[0] not in 'rwxa':
        raise ValueError("mode must start with 'r', 'w', 'a' or 'x'")
    if not _valid_chars.issuperset(mode):
        raise ValueError("mode '{}' contains invalid characters".format(mode))



from __future__ import unicode_literals

import os

from fs.tempfs import TempFS
from fs import errors
import mock

from .test_osfs import TestOSFS


class TestTempFS(TestOSFS):
    """Test OSFS implementation."""

    def make_fs(self):
        return TempFS()

    def test_clean(self):
        t = TempFS()
        _temp_dir = t.getsyspath('/')
        self.assertTrue(os.path.isdir(_temp_dir))
        t.close()
        self.assertFalse(os.path.isdir(_temp_dir))

    @mock.patch("shutil.rmtree", create=True)
    def test_clean_error(self, rmtree):
        rmtree.side_effect = Exception('boom')
        with self.assertRaises(errors.OperationFailed):
            t = TempFS(ignore_clean_errors=False)
            t.setbytes('foo', b'bar')
            t.close()



"""

Create a 'mirror' of a filesystem.

"""

from __future__ import print_function
from __future__ import unicode_literals


from .copy import copy_file
from .errors import ResourceNotFound
from .walk import Walker
from .opener import manage_fs


def _compare(info1, info2):
    """
    Compare two (file) info objects and return True if the file should
    be copied, or False if they should not.

    """
    # Check filesize has changed
    if info1.size != info2.size:
        return True
    # Check modified dates
    date1 = info1.modified
    date2 = info2.modified
    return date1 is None or date2 is None or date1 > date2


def mirror(src_fs, dst_fs, walker=None, copy_if_newer=True):
    """
    Mirror files / directories from one filesystem to another.

    :param src_fs: A source filesystem.
    :type src_fs: FS URL or instance.
    :param dst_fs: A destination filesystem.
    :type dst_fs: FS URL or instance.
    :param walker: An optional waler instance.
    :type walker: :class:`~fs.walk.Walker`
    :param bool copy_if_newer: Only copy newer files.

    Mirroring a filesystem will create an exact copy of ``src_fs`` on
    ``dst_fs``, by removing any files / directories on the destination
    that aren't on the source, and copying files that aren't.

    """
    with manage_fs(src_fs, writeable=False) as _src_fs:
        with manage_fs(dst_fs, create=True) as _dst_fs:
            with _src_fs.lock(), _dst_fs.lock():
                return _mirror(
                    _src_fs,
                    _dst_fs,
                    walker=walker,
                    copy_if_newer=copy_if_newer
                )


def _mirror(src_fs, dst_fs, walker=None, copy_if_newer=True):
    walker = walker or Walker()
    walk = walker.walk(src_fs, namespaces=['details'])
    for path, dirs, files in walk:
        try:
            dst = {
                info.name: info
                for info in dst_fs.scandir(path, namespaces=['details'])
            }
        except ResourceNotFound:
            dst_fs.makedir(path)
            dst = {}

        # Copy files
        for _file in files:
            _path = _file.make_path(path)
            dst_file = dst.pop(_file.name, None)
            if dst_file is not None:
                if dst_file.is_dir:
                    # Destination is a directory, remove it
                    dst_fs.removetree(_path)
                else:
                    # Compare file info
                    if copy_if_newer and not _compare(_file, dst_file):
                        continue
            copy_file(src_fs, _path, dst_fs, _path)

        # Make directories
        for _dir in dirs:
            _path = _dir.make_path(path)
            dst_dir = dst.pop(_dir.name, None)
            if dst_dir is not None:
                # Directory name exists on dst
                if not dst_dir.is_dir:
                    # Not a directory, so remove it
                    dst_fs.remove(_path)
            else:
                # Make the directory in dst
                dst_fs.makedir(_path, recreate=True)

        # Remove any remaining resources
        while dst:
            _, info = dst.popitem()
            _path = info.make_path(path)
            if info.is_dir:
                dst_fs.removetree(_path)
            else:
                dst_fs.remove(_path)



from __future__ import unicode_literals

import unittest

from six import text_type

from fs import errors


class TestErrors(unittest.TestCase):

    def test_str(self):
        err = errors.FSError('oh dear')
        repr(err)
        self.assertEqual(
            text_type(err),
            'oh dear'
        )

    def test_unsupported(self):
        err = errors.Unsupported('stuff')
        repr(err)
        self.assertEqual(
            text_type(err),
            "not supported"
        )



"""
Useful functions for working with PyFilesystem paths.
This is broadly similar to the standard ``os.path`` module but works
with paths in the canonical format expected by all FS objects (that is,
separated by forward slashes and with an optional leading slash).

See :ref:`paths` for an explanation of PyFilesystem paths.

"""

from __future__ import print_function
from __future__ import unicode_literals

import re

from .errors import IllegalBackReference


__all__ = [
    "abspath",
    "basename",
    "combine",
    "dirname",
    "forcedir",
    "frombase",
    "isabs",
    "isbase",
    "isdotfile",
    "isparent",
    "issamedir",
    "iswildcard",
    "iteratepath",
    "join",
    "normpath",
    "recursepath",
    "relativefrom",
    "relpath",
    "split",
    "splitext",
]

_requires_normalization = re.compile(
    r'(^|/)\.\.?($|/)|//',
    re.UNICODE
).search


def normpath(path):
    """
    Normalize a path.

    This function simplifies a path by collapsing back-references
    and removing duplicated separators.

    :param str path: Path to normalize.
    :returns: A valid FS path.
    :type: str

    >>> normpath("/foo//bar/frob/../baz")
    '/foo/bar/baz'
    >>> normpath("foo/../../bar")
    Traceback (most recent call last)
        ...
    IllegalBackReference: Too many backrefs in 'foo/../../bar'

    """
    if path in '/':
        return path

    # An early out if there is no need to normalize this path
    if not _requires_normalization(path):
        return path.rstrip('/')

    prefix = '/' if path.startswith('/') else ''
    components = []
    try:
        for component in path.split('/'):
            if component in '..':  # True for '..', '.', and ''
                if component == '..':
                    components.pop()
            else:
                components.append(component)
    except IndexError:
        raise IllegalBackReference(
            "Too many backrefs in '{}'".format(path)
        )
    return prefix + '/'.join(components)


def iteratepath(path):
    """
    Iterate over the individual components of a path.

    >>> iteratepath('/foo/bar/baz')
    ['foo', 'bar', 'baz']

    :param str path: Path to iterate over.
    :returns: A list of path components.
    :rtype: list

    """
    path = relpath(normpath(path))
    if not path:
        return []
    return path.split('/')


def recursepath(path, reverse=False):
    """
    Get intermediate paths from the root to the given path.

    :param str path: A PyFilesystem path
    :param bool reverse: Reverses the order of the paths.
    :returns: A list of paths.
    :rtype: list

    >>> recursepath('a/b/c')
    ['/', '/a', '/a/b', '/a/b/c']

    """
    if path in '/':
        return ['/']

    path = abspath(normpath(path)) + '/'

    paths = ['/']
    find = path.find
    append = paths.append
    pos = 1
    len_path = len(path)

    while pos < len_path:
        pos = find('/', pos)
        append(path[:pos])
        pos += 1

    if reverse:
        return paths[::-1]
    return paths


def isabs(path):
    """
    Check if a path is an absolute path.

    :param str path: A PyFilesytem path.
    :rtype: bool

    """
    # Somewhat trivial, but helps to make code self-documenting
    return path.startswith('/')


def abspath(path):
    """
    Convert the given path to an absolute path.

    Since FS objects have no concept of a *current directory*, this
    simply adds a leading ``/`` character if the path doesn't already
    have one.

    :param str path: A PyFilesytem path.
    :returns: An absolute path.
    :rtype: str

    """
    if not path.startswith('/'):
        return '/' + path
    return path


def relpath(path):
    """
    Convert the given path to a relative path.

    This is the inverse of abspath(), stripping a leading ``'/'`` from
    the path if it is present.

    :param str path: Path to adjust
    :rtype: str

    >>> relpath('/a/b')
    'a/b'

    """
    return path.lstrip('/')


def join(*paths):
    """
    Join any number of paths together.

    :param paths: Paths to join are given in positional arguments.
    :rtype: str

    >>> join('foo', 'bar', 'baz')
    'foo/bar/baz'
    >>> join('foo/bar', '../baz')
    'foo/baz'
    >>> join('foo/bar', '/baz')
    '/baz'

    """
    absolute = False
    relpaths = []
    for p in paths:
        if p:
            if p[0] == '/':
                del relpaths[:]
                absolute = True
            relpaths.append(p)

    path = normpath("/".join(relpaths))
    if absolute:
        path = abspath(path)
    return path


def combine(path1, path2):
    """
    Join two paths together.

    :param str path1: A PyFilesytem path.
    :param str path2: A PyFilesytem path.
    :rtype: str

    This is faster than :func:`~fs.path.join`, but only works when the
    second path is relative, and there are no back references in either
    path.

    >>> combine("foo/bar", "baz")
    'foo/bar/baz'

    """
    if not path1:
        return path2.lstrip()
    return "{}/{}".format(path1.rstrip('/'), path2.lstrip('/'))


def split(path):
    """
    Split a path into (head, tail) pair.

    This function splits a path into a pair (head, tail) where 'tail' is
    the last pathname component and 'head' is all preceding components.

    :param str path: Path to split
    :returns: tuple of ``(head, tail)``
    :rtype: tuple

    >>> split("foo/bar")
    ('foo', 'bar')
    >>> split("foo/bar/baz")
    ('foo/bar', 'baz')
    >>> split("/foo/bar/baz")
    ('/foo/bar', 'baz')

    """
    if '/' not in path:
        return ('', path)
    split = path.rsplit('/', 1)
    return (split[0] or '/', split[1])


def splitext(path):
    """
    Split the extension from the path, and returns the path (up to the
    last '.' and the extension).

    :param path: A path to split
    :returns: tuple of ``(path, extension)``
    :rtype: tuple

    >>> splitext('baz.txt')
    ('baz', '.txt')
    >>> splitext('foo/bar/baz.txt')
    ('foo/bar/baz', '.txt')

    """

    parent_path, pathname = split(path)
    if '.' not in pathname:
        return path, ''
    pathname, ext = pathname.rsplit('.', 1)
    path = join(parent_path, pathname)
    return path, '.' + ext


def isdotfile(path):
    """
    Detect if a path references a dot file, i.e. a resource who's name
    starts with a '.'

    :param path: Path to check.
    :type path: str
    :rtype: bool

    >>> isdotfile('.baz')
    True
    >>> isdotfile('foo/bar/.baz')
    True
    >>> isdotfile('foo/bar.baz')
    False

    """
    return basename(path).startswith('.')


def dirname(path):
    """
    Return the parent directory of a path.

    This is always equivalent to the 'head' component of the value
    returned by ``split(path)``.

    :param str path: A PyFilesytem path.
    :rtype: str

    >>> dirname('foo/bar/baz')
    'foo/bar'
    >>> dirname('/foo/bar')
    '/foo'
    >>> dirname('/foo')
    '/'

    """
    return split(path)[0]


def basename(path):
    """
    Return the basename of the resource referenced by a path.

    This is always equivalent to the 'tail' component of the value
    returned by split(path).

    :param str path: A PyFilesytem path.
    :rtype: str

    >>> basename('foo/bar/baz')
    'baz'
    >>> basename('foo/bar')
    'bar'
    >>> basename('foo/bar/')
    ''

    """
    return split(path)[1]


def issamedir(path1, path2):
    """
    Check if two paths reference a resource in the same directory.

    :param str path1: A PyFilesytem path.
    :param str path2: A PyFilesytem path.
    :rtype: bool

    >>> issamedir("foo/bar/baz.txt", "foo/bar/spam.txt")
    True
    >>> issamedir("foo/bar/baz/txt", "spam/eggs/spam.txt")
    False

    """
    return dirname(normpath(path1)) == dirname(normpath(path2))


def isbase(path1, path2):
    """
    Check if path1 is a base of path2.

    :param str path1: A PyFilesytem path.
    :param str path2: A PyFilesytem path.
    :rtype: bool

    """
    _path1 = forcedir(abspath(path1))
    _path2 = forcedir(abspath(path2))
    return _path1.startswith(_path2)


def isparent(path1, path2):
    """
    Check if ``path1`` is a parent directory of ``path2``.

    :param str path1: A PyFilesytem path.
    :param str path2: A PyFilesytem path.
    :rtype: bool

    >>> isparent("foo/bar", "foo/bar/spam.txt")
    True
    >>> isparent("foo/bar/", "foo/bar")
    True
    >>> isparent("foo/barry", "foo/baz/bar")
    False
    >>> isparent("foo/bar/baz/", "foo/baz/bar")
    False

    """
    bits1 = path1.split("/")
    bits2 = path2.split("/")
    while bits1 and bits1[-1] == "":
        bits1.pop()
    if len(bits1) > len(bits2):
        return False
    for (bit1, bit2) in zip(bits1, bits2):
        if bit1 != bit2:
            return False
    return True


def forcedir(path):
    """
    Ensure the path ends with a trailing forward slash

    :param path: A PyFilesytem path.
    :rtype: bool

    >>> forcedir("foo/bar")
    'foo/bar/'
    >>> forcedir("foo/bar/")
    'foo/bar/'

    """

    if not path.endswith('/'):
        return path + '/'
    return path


def frombase(path1, path2):
    """
    Get the final path of ``path2`` that isn't in ``path1``.

    :param str path1: A PyFilesytem path.
    :param str path2: A PyFilesytem path.
    :rtype: str

    >>> frombase('foo/bar/', 'foo/bar/baz/egg')
    'baz/egg'

    """
    if not isparent(path1, path2):
        raise ValueError("path1 must be a prefix of path2")
    return path2[len(path1):]


def relativefrom(base, path):
    """
    Return a path relative from a given base path, i.e. insert backrefs
    as appropriate to reach the path from the base.

    :param str base: Path to a directory.
    :param atr path: Path you wish to make relative.

    >>> relativefrom("foo/bar", "baz/index.html")
    '../../baz/index.html'

    """
    base = list(iteratepath(base))
    path = list(iteratepath(path))

    common = 0
    for component_a, component_b in zip(base, path):
        if component_a != component_b:
            break
        common += 1

    return '/'.join(['..'] * (len(base) - common) + path[common:])


_WILD_CHARS = frozenset('*?[]!{}')


def iswildcard(path):
    """
    Check if a path ends with a wildcard.

    :param int path: An FS path.
    :rtype: bool

    >>> iswildcard('foo/bar/baz.*')
    True
    >>> iswildcard('foo/bar')
    False

    """
    assert path is not None
    return not _WILD_CHARS.isdisjoint(path)



# coding: utf-8
"""
fs.opener
========

Imported at the same time as PyFilesystem2, contains
various objects and functions to open and manage FS.
"""

# Declare fs.opener as a namespace package
__import__('pkg_resources').declare_namespace(__name__)

# Import objects into fs.opener namespace
from .base import Opener
from .parse import parse_fs_url as parse
from .registry import registry

# Alias functions defined as Registry methods
open_fs = registry.open_fs
open = registry.open
manage_fs = registry.manage_fs

# __all__ with aliases and classes
__all__ = [
    "registry",
    "Opener",
    'open_fs',
    'open',
    'manage_fs',
    'parse',
]



from __future__ import unicode_literals, print_function

"""
  fstests.test_path:  testcases for the fs path functions

"""


import unittest

from fs.path import *


class TestPathFunctions(unittest.TestCase):
    """Testcases for FS path functions."""

    def test_normpath(self):
        tests = [
            ("\\a\\b\\c", "\\a\\b\\c"),
            (".", ""),
            ("./", ""),
            ("", ""),
            ("/.", "/"),
            ("/a/b/c", "/a/b/c"),
            ("a/b/c", "a/b/c"),
            ("a/b/../c/", "a/c"),
            ("/", "/"),
            ("a/\N{GREEK SMALL LETTER BETA}/c", "a/\N{GREEK SMALL LETTER BETA}/c"),
        ]
        for path, result in tests:
            self.assertEqual(normpath(path), result)

    def test_pathjoin(self):
        tests = [
            ("", "a", "a"),
            ("a", "a", "a/a"),
            ("a/b", "../c", "a/c"),
            ("a/b/../c", "d", "a/c/d"),
            ("/a/b/c", "d", "/a/b/c/d"),
            ("/a/b/c", "../../../d", "/d"),
            ("a", "b", "c", "a/b/c"),
            ("a/b/c", "../d", "c", "a/b/d/c"),
            ("a/b/c", "../d", "/a", "/a"),
            ("aaa", "bbb/ccc", "aaa/bbb/ccc"),
            ("aaa", "bbb\\ccc", "aaa/bbb\\ccc"),
            ("aaa", "bbb", "ccc", "/aaa", "eee", "/aaa/eee"),
            ("a/b", "./d", "e", "a/b/d/e"),
            ("/", "/", "/"),
            ("/", "", "/"),
            (u"a/\N{GREEK SMALL LETTER BETA}", "c", u"a/\N{GREEK SMALL LETTER BETA}/c"),
        ]
        for testpaths in tests:
            paths = testpaths[:-1]
            result = testpaths[-1]
            self.assertEqual(join(*paths), result)

        self.assertRaises(ValueError, join, "..")
        self.assertRaises(ValueError, join, "../")
        self.assertRaises(ValueError, join, "/..")
        self.assertRaises(ValueError, join, "./../")
        self.assertRaises(ValueError, join, "a/b", "../../..")
        self.assertRaises(ValueError, join, "a/b/../../../d")

    def test_relpath(self):
        tests = [
            ("/a/b", "a/b"),
            ("a/b", "a/b"),
            ("/", "")
        ]
        for path, result in tests:
            self.assertEqual(relpath(path), result)

    def test_abspath(self):
        tests = [
            ("/a/b", "/a/b"),
            ("a/b", "/a/b"),
            ("/", "/")
        ]
        for path, result in tests:
            self.assertEqual(abspath(path), result)

    def test_forcedir(self):
        self.assertEqual(forcedir('foo'), 'foo/')
        self.assertEqual(forcedir('foo/'), 'foo/')

    def test_frombase(self):
        with self.assertRaises(ValueError):
            frombase('foo', 'bar/baz')
        self.assertEqual(frombase('foo', 'foo/bar'), '/bar')

    def test_isabs(self):
        self.assertTrue(isabs('/'))
        self.assertTrue(isabs('/foo'))
        self.assertFalse(isabs('foo'))

    def test_iteratepath(self):
        tests = [
            ("a/b", ["a", "b"]),
            ("", []),
            ("aaa/bbb/ccc", ["aaa", "bbb", "ccc"]),
            ("a/b/c/../d", ["a", "b", "d"])
        ]

        for path, results in tests:
            for path_component, expected in zip(iteratepath(path), results):
                self.assertEqual(path_component, expected)

    def test_combine(self):
        self.assertEqual(combine('', 'bar'), 'bar')
        self.assertEqual(combine('foo', 'bar'), 'foo/bar')

    def test_pathsplit(self):
        tests = [
            ("a/b", ("a", "b")),
            ("a/b/c", ("a/b", "c")),
            ("a", ("", "a")),
            ("", ("", "")),
            ("/", ("/", "")),
            ("/foo", ("/", "foo")),
            ("foo/bar", ("foo", "bar")),
            ("foo/bar/baz", ("foo/bar", "baz")),
        ]
        for path, result in tests:
            self.assertEqual(split(path), result)

    def test_splitext(self):
        self.assertEqual(splitext('foo.bar'), ('foo', '.bar'))
        self.assertEqual(splitext('foo.bar.baz'), ('foo.bar', '.baz'))
        self.assertEqual(splitext('foo'), ('foo', ''))

    def test_recursepath(self):
        self.assertEquals(recursepath("/"), ["/"])
        self.assertEquals(recursepath("hello"), ["/", "/hello"])
        self.assertEquals(recursepath("/hello/world/"), ["/", "/hello", "/hello/world"])
        self.assertEquals(recursepath("/hello/world/", reverse=True),
                          ["/hello/world", "/hello", "/"])
        self.assertEquals(recursepath("hello", reverse=True), ["/hello", "/"])
        self.assertEquals(recursepath("", reverse=True), ["/"])

    def test_isbase(self):
        self.assertTrue(isbase('foo/bar', 'foo'))
        self.assertFalse(isbase('foo/bar', 'baz'))

    def test_isparent(self):
        self.assertTrue(isparent("foo/bar", "foo/bar/spam.txt"))
        self.assertTrue(isparent("foo/bar/", "foo/bar"))
        self.assertFalse(isparent("foo/barry", "foo/baz/bar"))
        self.assertFalse(isparent("foo/bar/baz/", "foo/baz/bar"))
        self.assertFalse(isparent("foo/var/baz/egg", "foo/baz/bar"))

    def test_issamedir(self):
        self.assertTrue(issamedir("foo/bar/baz.txt", "foo/bar/spam.txt"))
        self.assertFalse(issamedir("foo/bar/baz/txt", "spam/eggs/spam.txt"))

    def test_isdotfile(self):
        for path in ['.foo',
                     '.svn',
                     'foo/.svn',
                     'foo/bar/.svn',
                     '/foo/.bar']:
            self.assert_(isdotfile(path))

        for path in ['asfoo',
                     'df.svn',
                     'foo/er.svn',
                     'foo/bar/test.txt',
                     '/foo/bar']:
            self.assertFalse(isdotfile(path))

    def test_dirname(self):
        tests = [('foo', ''),
                 ('foo/bar', 'foo'),
                 ('foo/bar/baz', 'foo/bar'),
                 ('/foo/bar', '/foo'),
                 ('/foo', '/'),
                 ('/', '/')]
        for path, test_dirname in tests:
            self.assertEqual(dirname(path), test_dirname)

    def test_basename(self):
        tests = [('foo', 'foo'),
                 ('foo/bar', 'bar'),
                 ('foo/bar/baz', 'baz'),
                 ('/', '')]
        for path, test_basename in tests:
            self.assertEqual(basename(path), test_basename)

    def test_iswildcard(self):
        self.assert_(iswildcard('*'))
        self.assert_(iswildcard('*.jpg'))
        self.assert_(iswildcard('foo/*'))
        self.assert_(iswildcard('foo/{}'))
        self.assertFalse(iswildcard('foo'))
        self.assertFalse(iswildcard('img.jpg'))
        self.assertFalse(iswildcard('foo/bar'))

    def test_realtivefrom(self):
        tests = [('/', '/foo.html', 'foo.html'),
                 ('/foo', '/foo/bar.html', 'bar.html'),
                 ('/foo/bar/', '/egg.html', '../../egg.html'),
                 ('/a/b/c/d', 'e', '../../../../e'),
                 ('/a/b/c/d', 'a/d', '../../../d'),
                 ('/docs/', 'tags/index.html', '../tags/index.html'),
                 ('foo/bar', 'baz/index.html', '../../baz/index.html'),
                 ('', 'a', 'a'),
                 ('a', 'b/c', '../b/c')
                 ]

        for base, path, result in tests:
            self.assertEqual(relativefrom(base, path), result)



from __future__ import unicode_literals

import os
import six
import gzip
import tarfile
import tempfile
import unittest

from fs import tarfs
from fs import errors
from fs.compress import write_tar
from fs.opener import open_fs
from fs.test import FSTestCases

from .test_archives import ArchiveTestCases


class TestWriteTarFS(FSTestCases, unittest.TestCase):
    """
    Test TarFS implementation.

    When writing, a TarFS is essentially a TempFS.
    """

    def make_fs(self):
        fh, _tar_file = tempfile.mkstemp()
        os.close(fh)
        fs = tarfs.TarFS(_tar_file, write=True)
        fs._tar_file = _tar_file
        return fs

    def destroy_fs(self, fs):
        fs.close()
        del fs._tar_file

class TestWriteTarFSToFileobj(FSTestCases, unittest.TestCase):
    """
    Test TarFS implementation.

    When writing, a TarFS is essentially a TempFS.

    """

    def make_fs(self):
        _tar_file = six.BytesIO()
        fs = tarfs.TarFS(_tar_file, write=True)
        fs._tar_file = _tar_file
        return fs

    def destroy_fs(self, fs):
        fs.close()
        del fs._tar_file

class TestWriteGZippedTarFS(FSTestCases, unittest.TestCase):

    def make_fs(self):
        fh, _tar_file = tempfile.mkstemp()
        os.close(fh)
        fs = tarfs.TarFS(_tar_file, write=True, compression="gz")
        fs._tar_file = _tar_file
        return fs

    def destroy_fs(self, fs):
        fs.close()
        del fs._tar_file

    def assert_is_bzip(self):
        try:
            tarfile.open(fs._tar_file, 'r:gz')
        except tarfile.ReadError:
            self.fail("{} is not a valid gz archive".format(fs._tar_file))
        for other_comps in ['xz', 'bz2', '']:
            with self.assertRaises(tarfile.ReadError):
                tarfile.open(fs._tar_file,
                             'r:{}'.format(other_comps))

@unittest.skipIf(six.PY2, "Python2 does not support LZMA")
class TestWriteXZippedTarFS(FSTestCases, unittest.TestCase):

    def make_fs(self):
        fh, _tar_file = tempfile.mkstemp()
        os.close(fh)
        fs = tarfs.TarFS(_tar_file, write=True, compression="xz")
        fs._tar_file = _tar_file
        return fs

    def destroy_fs(self, fs):
        fs.close()
        self.assert_is_xz(fs)
        os.remove(fs._tar_file)
        del fs._tar_file

    def assert_is_xz(self, fs):
        try:
            tarfile.open(fs._tar_file, 'r:xz')
        except tarfile.ReadError:
            self.fail("{} is not a valid xz archive".format(fs._tar_file))
        for other_comps in ['bz2', 'gz', '']:
            with self.assertRaises(tarfile.ReadError):
                tarfile.open(fs._tar_file,
                             'r:{}'.format(other_comps))

class TestWriteBZippedTarFS(FSTestCases, unittest.TestCase):

    def make_fs(self):
        fh, _tar_file = tempfile.mkstemp()
        os.close(fh)
        fs = tarfs.TarFS(_tar_file, write=True, compression="bz2")
        fs._tar_file = _tar_file
        return fs

    def destroy_fs(self, fs):
        fs.close()
        self.assert_is_bzip(fs)
        os.remove(fs._tar_file)
        del fs._tar_file

    def assert_is_bzip(self, fs):
        try:
            tarfile.open(fs._tar_file, 'r:bz2')
        except tarfile.ReadError:
            self.fail("{} is not a valid bz2 archive".format(fs._tar_file))
        for other_comps in ['gz', '']:
            with self.assertRaises(tarfile.ReadError):
                tarfile.open(fs._tar_file,
                             'r:{}'.format(other_comps))

class TestReadTarFS(ArchiveTestCases, unittest.TestCase):
    """
    Test Reading tar files.

    """
    def compress(self, fs):
        fh, self._temp_path = tempfile.mkstemp()
        os.close(fh)
        write_tar(fs, self._temp_path)

    def load_archive(self):
        return tarfs.TarFS(self._temp_path)

    def remove_archive(self):
        os.remove(self._temp_path)

    def test_read_from_fileobject(self):
        try:
            tarfs.TarFS(open(self._temp_path, 'rb'))
        except:
            self.fail("Couldn't open tarfs from fileobject")

    def test_read_from_filename(self):
        try:
            tarfs.TarFS(self._temp_path)
        except:
            self.fail("Couldn't open tarfs from fileobject")

class TestReadTarFSMem(TestReadTarFS):

    def make_source_fs(self):
        return open_fs('mem://')



"""

The machinery for walking a filesystem. See :ref:`walking` for details.

"""

from __future__ import unicode_literals

import itertools

from collections import deque
from collections import namedtuple

import six

from ._repr import make_repr
from .errors import FSError
from .path import abspath
from .path import join
from .path import normpath


"""A 'step' in a directory walk."""
Step = namedtuple('Step', 'path, dirs, files')


class WalkerBase(object):
    """
    The base class for a Walker.

    To create a custom walker, implement
    :meth:`~fs.walk.WalkerBase.walk` in a sub-class.

    See :meth:`~fs.walk.Walker` for a fully featured walker object that
    should be adequate for all but your most exotic directory walking
    needs.

    """

    def walk(self, fs, path='/', namespaces=None):
        """
        Walk the directory structure of a filesystem.

        :param fs: A FS object.
        :param str path: a path to a directory.
        :param list namespaces: A list of additional namespaces to add
            to the Info objects.
        :returns: Iterator of :class:`~fs.walk.Step` named tuples.

        """
        raise NotImplementedError

    def files(self, fs, path='/'):
        """
        Walk a filesystem, yielding absolute paths to files.

        :param fs: A FS object.
        :param str path: A path to a directory.
        :returns: An iterable of file paths.

        """
        iter_walk = iter(self.walk(fs, path=path))
        for _path, _, files in iter_walk:
            for info in files:
                yield join(_path, info.name)

    def dirs(self, fs, path='/'):
        """
        Walk a filesystem, yielding absolute paths to directories.

        :param str fs: A FS object.
        :param str path: A path to a directory.

        """
        for path, dirs, _ in self.walk(fs, path=path):
            for info in dirs:
                yield join(path, info.name)

    def info(self, fs, path='/', namespaces=None):
        """
        Walk a filesystem, yielding tuples of ``(<absolute path>,
        <resource info>)``.

        :param str fs: A FS object.
        :param str path: A path to a directory.
        :param list namespaces: A list of additional namespaces to add
            to the Info objects.
        :returns: An iterable of :class:`~fs.info.Info` objects.

        """
        _walk = self.walk(fs, path=path, namespaces=namespaces)
        for _path, dirs, files in _walk:
            for info in itertools.chain(dirs, files):
                yield join(_path, info.name), info


class Walker(WalkerBase):
    """
    A walker object recursively lists directories in a filesystem.

    :param bool ignore_errors: If true, any errors reading a
        directory will be ignored, otherwise exceptions will be
        raised.
    :param callable on_error: If ``ignore_errors`` is false, then
        this callable will be invoked with a path and the exception
        object. It should return True to ignore the error, or False
        to re-raise it.
    :param str search: If ``'breadth'`` then the directory will be
        walked *top down*. Set to ``'depth'`` to walk *bottom up*.
    :param list filter: If supplied, this parameter should be a list of
        filename patterns, e.g. ``['*.py']``. Files will only be
        returned if the final component matches one of the patterns.
    :param list exclude_dirs: A list of patterns that will be used
        to filter out directories from the walk. e.g. ``['*.svn',
        '*.git']``.

    """

    def __init__(self,
                 ignore_errors=False,
                 on_error=None,
                 search="breadth",
                 filter=None,
                 exclude_dirs=None):
        if search not in ('breadth', 'depth'):
            raise ValueError("search must be 'breadth' or 'depth'")
        self.ignore_errors = ignore_errors
        if ignore_errors:
            assert on_error is None,\
                'on_error is invalid when ignore_errors==True'
            on_error = lambda path, error: True
        self.on_error = on_error or (lambda path, error: True)
        self.search = search
        self.filter = filter
        self.exclude_dirs = exclude_dirs
        super(Walker, self).__init__()

    @classmethod
    def bind(cls, fs):
        """
        This *binds* in instance of the Walker to a given filesystem, so
        that you won't need to explicitly provide the filesystem as a
        parameter. Here's an example of binding::

            >>> from fs import open_fs
            >>> from fs.walk import Walker
            >>> home_fs = open_fs('~/')
            >>> walker = Walker.bind(home_fs)
            >>> for path in walker.files(filter=['*.py']):
            ...     print(path)

        Unless you have written a customized walker class, you will be
        unlikely to need to call this explicitly, as filesystem objects
        already have a ``walk`` attribute which is a bound walker
        object. Here's how you might use it::

            >>> from fs import open_fs
            >>> home_fs = open_fs('~/')
            >>> for path in home_fs.walk.files(filter=['*.py']):
            ...     print(path)

        :param fs: A filesystem object.
        :returns: a :class:`~fs.walk.BoundWalker`

        """
        return BoundWalker(fs)

    def __repr__(self):
        return make_repr(
            self.__class__.__name__,
            ignore_errors=(self.ignore_errors, False),
            on_error=(self.on_error, None),
            search=(self.search, 'breadth'),
            filter=(self.filter, None),
            exclude_dirs=(self.exclude_dirs, None)
        )

    def filter_files(self, fs, infos):
        """
        Filters a sequence of resource Info objects.

        The default implementation filters those files for which
        :meth:`~fs.walk.Walker.check_file` returns True.

        :param fs: A FS object.
        :type fs: :class:`~fs.base.FS`
        :param infos: A list of :class:`~fs.info.Info` instances.
        :type infos: list
        :rtype: list

        """
        _check_file = self.check_file
        return [
            info for info in infos
            if _check_file(fs, info)
        ]


    def check_open_dir(self, fs, info):
        """
        Check if a directory should be opened. Override to exclude
        directories from the walk.

        :param fs: A FS object.
        :type fs: :class:`~fs.base.FS`
        :param info: A :class:`~fs.info.Info` object.
        :rtype: bool

        """
        if self.exclude_dirs is None:
            return True
        return not fs.match(self.exclude_dirs, info.name)

    def check_file(self, fs, info):
        """
        Check if a filename should be included in the walk. Override to
        exclude files from the walk.

        :param fs: A FS object.
        :type fs: :class:`~fs.base.FS`
        :param info: A :class:`~fs.info.Info` object.
        :rtype: bool

        """
        return fs.match(self.filter, info.name)

    def _scan(self, fs, dir_path, namespaces):
        """
        Get an iterator of :class:`~fs.info.Info` objects for a
        directory path.

        :param fs: A FS object.
        :type fs: :class:`~fs.base.FS`
        :param str dir_path: A path to a directory.

        """
        try:
            return fs.scandir(dir_path, namespaces=namespaces)
        except FSError as error:
            if self.on_error(dir_path, error):
                return iter(())
            six.reraise(type(error), error)

    def walk(self, fs, path='/', namespaces=None):
        """
        Walk the directory structure of a filesystem.

        :param fs: A FS object.
        :param str path: a path to a directory.
        :param list namespaces: A list of additional namespaces to add
            to the Info objects.
        :returns: :class:`~fs.walk.Step` iterator.

        The return value is an iterator of ``(<path>, <dirs>, <files>)``
        named tuples,  where ``<path>`` is an absolute path to a
        directory, and ``<dirs>`` and ``<files>`` are a list of
        :class:`~fs.info.Info` objects for directories and files
        in ``<path>``.

        Here's an example::

            home_fs = open_fs('~/')
            walker = Walker(filter=['*.py'])
            for path, dirs, files in walker.walk(home_fs, namespaces=['details']):
                print("[{}]".format(path))
                print("{} directories".format(len(dirs)))
                total = sum(info.size for info in files)
                print("{} bytes {}".format(total))

        """
        _path = abspath(normpath(path))

        if self.search == 'breadth':
            do_walk = self._walk_breadth
        elif self.search == 'depth':
            do_walk = self._walk_depth

        return do_walk(fs, _path, namespaces=namespaces)

    def _walk_breadth(self, fs, path, namespaces=None):
        """
        Walk files using a breadth first search.

        """
        queue = deque([path])
        push = queue.appendleft
        pop = queue.pop

        while queue:
            dir_path = pop()
            dirs = []
            files = []
            for info in self._scan(fs, dir_path, namespaces=namespaces):
                if info.is_dir:
                    if self.check_open_dir(fs, info):
                        dirs.append(info)
                        push(join(dir_path, info.name))
                else:
                    files.append(info)
            yield (
                dir_path,
                dirs,
                self.filter_files(fs, files)
            )

    def _walk_depth(self, fs, path, namespaces=None):
        """
        Walk files using a depth first search.

        """
        # No recursion!

        def scan(path):
            return self._scan(fs, path, namespaces=namespaces)

        stack = [(
            path, scan(path), [], []
        )]
        push = stack.append

        while stack:
            dir_path, iter_files, dirs, files = stack[-1]
            try:
                info = next(iter_files)
            except StopIteration:
                yield (
                    dir_path,
                    dirs,
                    self.filter_files(fs, files)
                )
                del stack[-1]
            else:
                if info.is_dir:
                    if self.check_open_dir(fs, info):
                        dirs.append(info)
                        _path = join(dir_path, info.name)
                        push((
                            _path, scan(_path), [], []
                        ))
                else:
                    files.append(info)


class BoundWalker(object):
    """
    A class that binds a :class:`~fs.walk.Walker` instance to a FS
    object.

    :param fs: A FS object.
    :param walker_class: A :class:`~fs.walk.WalkerBase` sub-class. The
        default uses :class:`~fs.walk.Walker`.

    You will typically not need to create instances of this class
    explicitly. Filesystems have a ``walk`` property which returns a
    ``BoundWalker`` object.

        >>> import fs
        >>> home_fs = fs.open_fs('~/')
        >>> home_fs.walk
        BoundWalker(OSFS('/Users/will', encoding='utf-8'))

    A BoundWalker is callable. Calling it is an alias for
    :meth:`~fs.walk.BoundWalker.walk`.

    """

    def __init__(self, fs, walker_class=Walker):
        self.fs = fs
        self.walker_class = walker_class

    def __repr__(self):
        return "BoundWalker({!r})".format(self.fs)

    def _make_walker(self, *args, **kwargs):
        """Create a walker instance."""
        walker = self.walker_class(*args, **kwargs)
        return walker

    def walk(self,
             path='/',
             namespaces=None,
             **kwargs):
        """
        Walk the directory structure of a filesystem.

        :param str path: A path to a directory.
        :param bool ignore_errors: If true, any errors reading a
            directory will be ignored, otherwise exceptions will be
            raised.
        :param callable on_error: If ``ignore_errors`` is false, then
            this callable will be invoked with a path and the exception
            object. It should return True to ignore the error, or False
            to re-raise it.
        :param str search: If ``'breadth'`` then the directory will be
            walked *top down*. Set to ``'depth'`` to walk *bottom up*.
        :param list filter: If supplied, this parameter should be a list
            of file name patterns, e.g. ``['*.py']``. Files will only be
            returned if the final component matches one of the
            patterns.
        :param list exclude_dirs: A list of patterns that will be used
            to filter out directories from the walk, e.g. ``['*.svn',
            '*.git']``.
        :returns: :class:`~fs.walk.Step` iterator.

        The return value is an iterator of ``(<path>, <dirs>, <files>)``
        named tuples,  where ``<path>`` is an absolute path to a
        directory, and ``<dirs>`` and ``<files>`` are a list of
        :class:`~fs.info.Info` objects for directories and files
        in ``<path>``.

        Here's an example::

            home_fs = open_fs('~/')
            walker = Walker(filter=['*.py'])
            for path, dirs, files in walker.walk(home_fs, namespaces=['details']):
                print("[{}]".format(path))
                print("{} directories".format(len(dirs)))
                total = sum(info.size for info in files)
                print("{} bytes {}".format(total))

        This method invokes :meth:`~fs.walk.Walker.walk` with bound FS
        object.

        """
        walker = self._make_walker(**kwargs)
        return walker.walk(self.fs, path=path, namespaces=namespaces)

    __call__ = walk

    def files(self, path='/', **kwargs):
        """
        Walk a filesystem, yielding absolute paths to files.

        :param str path: A path to a directory.
        :param bool ignore_errors: If true, any errors reading a
            directory will be ignored, otherwise exceptions will be
            raised.
        :param callable on_error: If ``ignore_errors`` is false, then
            this callable will be invoked with a path and the exception
            object. It should return True to ignore the error, or False
            to re-raise it.
        :param str search: If ``'breadth'`` then the directory will be
            walked *top down*. Set to ``'depth'`` to walk *bottom up*.
        :param list filter: If supplied, this parameter should be a list
            of file name patterns, e.g. ``['*.py']``. Files will only be
            returned if the final component matches one of the
            patterns.
        :param list exclude_dirs: A list of patterns that will be used
            to filter out directories from the walk, e.g. ``['*.svn',
            '*.git']``.
        :returns: An iterable of file paths (absolute from the
            filesystem root).

        This method invokes :meth:`~fs.walk.Walker.files` with the bound
        FS object.

        """
        walker = self._make_walker(**kwargs)
        return walker.files(self.fs, path=path)

    def dirs(self, path='/', **kwargs):
        """
        Walk a filesystem, yielding absolute paths to directories.

        :param str path: A path to a directory.
        :param bool ignore_errors: If true, any errors reading a
            directory will be ignored, otherwise exceptions will be
            raised.
        :param callable on_error: If ``ignore_errors`` is false, then
            this callable will be invoked with a path and the exception
            object. It should return True to ignore the error, or False
            to re-raise it.
        :param str search: If ``'breadth'`` then the directory will be
            walked *top down*. Set to ``'depth'`` to walk *bottom up*.
        :param list exclude_dirs: A list of patterns that will be used
            to filter out directories from the walk, e.g. ``['*.svn',
            '*.git']``.
        :returns: An iterable of directory paths (absolute from the FS
            root).

        This method invokes :meth:`~fs.walk.Walker.dirs` with the bound
        FS object.

        """
        walker = self._make_walker(**kwargs)
        return walker.dirs(self.fs, path=path)

    def info(self, path='/', namespaces=None, **kwargs):
        """
        Walk a filesystem, yielding tuples of ``(<absolute path>,
        <resource info>)``.

        :param str path: A path to a directory.
        :param bool ignore_errors: If true, any errors reading a
            directory will be ignored, otherwise exceptions will be
            raised.
        :param callable on_error: If ``ignore_errors`` is false, then
            this callable will be invoked with a path and the exception
            object. It should return True to ignore the error, or False
            to re-raise it.
        :param str search: If ``'breadth'`` then the directory will be
            walked *top down*. Set to ``'depth'`` to walk *bottom up*.
        :param list filter: If supplied, this parameter should be a list
            of file name patterns, e.g. ``['*.py']``. Files will only be
            returned if the final component matches one of the
            patterns.
        :param list exclude_dirs: A list of patterns that will be used
            to filter out directories from the walk, e.g. ``['*.svn',
            '*.git']``.
        :returns: An iterable :class:`~fs.info.Info` objects.

        This method invokes :meth:`~fs.walk.Walker.info` with the bound
        FS object.

        """
        walker = self._make_walker(**kwargs)
        return walker.info(
            self.fs,
            path=path,
            namespaces=namespaces
        )

# Allow access to default walker from the module
# For example:
#     fs.walk.walk_files()

default_walker = Walker()
walk = default_walker.walk
walk_files = default_walker.files
walk_info = default_walker.info
walk_dirs = default_walker.dirs



"""
fs.wrap
=======

A collection of :class:`~fs.wrapfs.WrapFS` objects that modify the
behavior of another filesystem.

Here's an example that opens a filesystem then makes it *read only*::

    from fs import open_fs
    from fs.wrap import read_only

    projects_fs = open_fs('~/projects')
    read_only_projects_fs = read_only(projects_fs)

    # Will raise ResourceReadOnly exception
    read_only_projects_fs.remove('__init__.py')


"""

from __future__ import print_function
from __future__ import unicode_literals

from .wrapfs import WrapFS
from .path import abspath, normpath, split
from .errors import ResourceReadOnly, ResourceNotFound
from .info import Info
from .mode import check_writable


def read_only(fs):
    """
    Make a read-only filesystem.

    :param fs: A FS object.
    :returns: A read only version of ``fs``.

    """
    return WrapReadOnly(fs)


def cache_directory(fs):
    """
    Make a filesystem that caches directory information.

    :param fs: A FS object.
    :returns: A filesystem that caches results of ``scandir``, ``isdir``
        and other methods which read directory information.

    """
    return WrapCachedDir(fs)


class WrapCachedDir(WrapFS):
    """
    Caches filesystem directory information.

    This filesystem caches directory information retrieved from a
    scandir call. This *may* speed up code that calls ``isdir``,
    ``isfile``, or ``gettype`` too frequently.

    .. note::
        Using this wrap will prevent changes to directory information
        being visible to the filesystem object. Consequently it is best
        used only in a fairly limited scope where you don't expected
        anything on the filesystem to change.

    """

    wrap_name = 'cached-dir'

    def __init__(self, wrap_fs):
        super(WrapCachedDir, self).__init__(wrap_fs)
        self._cache = {}

    def scandir(self, path, namespaces=None, page=None):
        _path = abspath(normpath(path))
        cache_key = (_path, frozenset(namespaces or ()))
        if cache_key not in self._cache:
            _scan_result = self._wrap_fs.scandir(
                path,
                namespaces=namespaces,
                page=page
            )
            _dir = {info.name: info for info in _scan_result}
            self._cache[cache_key] = _dir
        gen_scandir = iter(self._cache[cache_key].values())
        return gen_scandir

    def getinfo(self, path, *namespaces):
        _path = abspath(normpath(path))
        if _path == '/':
            return Info({
                "basic": {
                    "name": "",
                    "is_dir": True
                }
            })
        dir_path, resource_name = split(_path)
        cache_key = (dir_path, frozenset(namespaces or ()))

        if cache_key not in self._cache:
            self.scandir(dir_path, namespaces=namespaces)

        _dir = self._cache[cache_key]
        try:
            info = _dir[resource_name]
        except KeyError:
            raise ResourceNotFound(path)
        return info

    def isdir(self, path):
        return self.getinfo(path).is_dir

    def isfile(self, path):
        return not self.getinfo(path).is_dir


class WrapReadOnly(WrapFS):
    """
    Makes a Filesystem read-only. Any call that would would write data
    or modify the filesystem in any way will raise a
    :class:`~fs.errors.ResourceReadOnly` exception.

    """

    wrap_name = 'read-only'

    def appendbytes(self, path, data):
        self.check()
        raise ResourceReadOnly(path)

    def appendtext(self, path, text,
                   encoding='utf-8', errors=None, newline=''):
        self.check()
        raise ResourceReadOnly(path)

    def makedir(self, path, permissions=None, recreate=False):
        self.check()
        raise ResourceReadOnly(path)

    def move(self, src_path, dst_path, overwrite=False):
        self.check()
        raise ResourceReadOnly(dst_path)

    def openbin(self, path, mode='r', buffering=-1, **options):
        self.check()
        if check_writable(mode):
            raise ResourceReadOnly(path)
        return self._wrap_fs.openbin(
            path,
            mode=mode,
            buffering=-1,
            **options
        )

    def remove(self, path):
        self.check()
        raise ResourceReadOnly(path)

    def removedir(self, path):
        self.check()
        raise ResourceReadOnly(path)

    def setinfo(self, path, info):
        self.check()
        raise ResourceReadOnly(path)

    def settext(self,
                path,
                contents,
                encoding='utf-8',
                errors=None,
                newline=''):
        self.check()
        raise ResourceReadOnly(path)

    def settimes(self, path, accessed=None, modified=None):
        self.check()
        raise ResourceReadOnly(path)

    def copy(self, src_path, dst_path, overwrite=False):
        self.check()
        raise ResourceReadOnly(dst_path)

    def create(self, path, wipe=False):
        self.check()
        raise ResourceReadOnly(path)

    def makedirs(self, path, recreate=False, mode=0o777):
        self.check()
        raise ResourceReadOnly(path)

    def open(self,
             path,
             mode='r',
             buffering=-1,
             encoding=None,
             errors=None,
             newline='',
             line_buffering=False,
             **options):
        self.check()
        if check_writable(mode):
            raise ResourceReadOnly(path)
        return self._wrap_fs.open(
            path,
            mode=mode,
            buffering=buffering,
            encoding=encoding,
            errors=errors,
            newline=newline,
            line_buffering=line_buffering,
            **options
        )

    def setbytes(self, path, contents):
        self.check()
        raise ResourceReadOnly(path)

    def setbinfile(self, path, file):
        self.check()
        raise ResourceReadOnly(path)

    def setfile(self,
                path,
                file,
                encoding=None,
                errors=None,
                newline=''):
        self.check()
        raise ResourceReadOnly(path)

    def touch(self, path):
        self.check()
        raise ResourceReadOnly(path)



from __future__ import print_function
from __future__ import unicode_literals

import tarfile
import six

from . import errors
from .base import FS
from .compress import write_tar
from .enums import ResourceType
from .info import Info
from .iotools import RawWrapper
from .opener import open_fs
from .path import dirname, normpath, relpath, basename
from .wrapfs import WrapFS
from .permissions import Permissions


class TarFS(WrapFS):
    """
    Read and write tar files.

    There are two ways to open a TarFS for the use cases of reading
    a tar file, and creating a new one.

    If you open the TarFS with  ``write`` set to ``False`` (the
    default), then the filesystem will be a read only filesystem which
    maps to the files and directories within the tar file. Files are
    decompressed on the fly when you open them.

    Here's how you might extract and print a readme from a tar file::

        with TarFS('foo.tar.gz') as tar_fs:
            readme = tar_fs.gettext('readme.txt')

    If you open the TarFS with ``write`` set to ``True``, then the TarFS
    will be a empty temporary filesystem. Any files / directories you
    create in the TarFS will be written in to a tar file when the TarFS
    is closed. The compression is set from the new file name but may be
    set manually with the ``compression`` argument.

    Here's how you might write a new tar file containing a readme.txt
    file::

        with TarFS('foo.tar.xz', write=True) as new_tar:
            new_tar.settext(
                'readme.txt',
                'This tar file was written by PyFilesystem'
            )

    :param file: An OS filename, or a open file object.
    :type file: str or file
    :param write: Set to ``True`` to write a new tar file, or ``False``
        to read an existing tar file.
    :type write: bool
    :param compression:  Compression to use (one of the formats
        supported by ``tarfile``: ``xz``, ``gz``, ``bz2``, or None).
    :type compression: str
    :param temp_fs: An opener string for the temporary filesystem
        used to store data prior to tarring.
    :type temp_fs: str

    """

    _compression_formats = {
        #FMT    #UNIX      #MSDOS
        'xz': ('.tar.xz', '.txz'),
        'bz2': ('.tar.bz2', '.tbz'),
        'gz': ('.tar.gz', '.tgz'),
    }

    def __new__(cls,
                file,
                write=False,
                compression=None,
                encoding="utf-8",
                temp_fs="temp://__tartemp__"):

        filename = str(getattr(file, 'name', file))

        if write and compression is None:
            compression = None
            for comp, extensions in six.iteritems(cls._compression_formats):
                if filename.endswith(extensions):
                    compression = comp
                    break

        if write:
            return WriteTarFS(file,
                              compression=compression,
                              encoding=encoding,
                              temp_fs=temp_fs)
        # else:
        return ReadTarFS(file, encoding=encoding)


@six.python_2_unicode_compatible
class WriteTarFS(WrapFS):
    """A writable tar file."""

    def __init__(self,
                 file,
                 compression=None,
                 encoding="utf-8",
                 temp_fs="temp://__tartemp__"):
        self._file = file
        self.compression = compression
        self.encoding = encoding
        self._temp_fs_url = temp_fs
        self._temp_fs = open_fs(temp_fs)
        self._meta = self._temp_fs.getmeta().copy()
        super(WriteTarFS, self).__init__(self._temp_fs)

    def __repr__(self):
        t = "WriteTarFS({!r}, compression={!r}, encoding={!r}, temp_fs={!r})"
        return t.format(
            self._file,
            self.compression,
            self.encoding,
            self._temp_fs_url
        )

    def __str__(self):
        return "<TarFS-write '{}'>".format(self._file)

    def delegate_path(self, path):
        return self._temp_fs, path

    def delegate_fs(self):
        return self._temp_fs

    def close(self):
        if not self.isclosed():
            try:
                self.write_tar()
            finally:
                self._temp_fs.close()
        super(WriteTarFS, self).close()

    def write_tar(self, file=None, compression=None, encoding=None):
        """
        Write tar to a file.

        .. note::
            This is called automatically when the TarFS is closed.

        :param file: Destination file, may be a file name or an open
            file object.
        :type file: str or file-like
        :param compression: Compression to use (one of the constants
            defined in the tarfile module in the stdlib).

        """
        if not self.isclosed():
            write_tar(
                self._temp_fs,
                file or self._file,
                compression=compression or self.compression,
                encoding=encoding or self.encoding
            )


@six.python_2_unicode_compatible
class ReadTarFS(FS):
    """A readable tar file."""

    _meta = {
        'case_insensitive': True,
        'network': False,
        'read_only': True,
        'supports_rename': False,
        'thread_safe': True,
        'unicode_paths': True,
        'virtual': False,
    }

    _typemap = type_map = {
        tarfile.BLKTYPE: ResourceType.block_special_file,
        tarfile.CHRTYPE: ResourceType.character,
        tarfile.DIRTYPE: ResourceType.directory,
        tarfile.FIFOTYPE: ResourceType.fifo,
        tarfile.REGTYPE: ResourceType.file,
        tarfile.AREGTYPE: ResourceType.file,
        tarfile.SYMTYPE: ResourceType.symlink,
        tarfile.CONTTYPE: ResourceType.file,
        tarfile.LNKTYPE: ResourceType.symlink,
    }

    def __init__(self, file, encoding='utf-8'):
        super(ReadTarFS, self).__init__()
        self._file = file
        self.encoding = encoding
        if hasattr(file, 'read'):
            self._tar = tarfile.open(fileobj=file, mode='r')
        else:
            self._tar = tarfile.open(file, mode='r')

    def __repr__(self):
        return "ReadTarFS({!r})".format(self._file)

    def __str__(self):
        return "<TarFS '{}'>".format(self._file)

    def getinfo(self, path, namespaces=None):
        self.check()
        namespaces = namespaces or ()
        path = relpath(normpath(path))
        if not path:
            raw_info = {
                "basic":
                {
                    "name": "",
                    "is_dir": True,
                },
                "details":
                {
                    "type": int(ResourceType.directory)
                }
            }
        else:
            try:
                member = self._tar.getmember(path)
            except KeyError:
                raise errors.ResourceNotFound(path)


            raw_tar_info = member.get_info(*(
                [self.encoding, None] if six.PY2 else []
            ))

            raw_tar_info.update({
                k.replace('is', 'is_'):getattr(member, k)()
                for k in dir(member)
                if k.startswith('is')
            })
            raw_info = {
                "basic":
                {
                    "name": basename(member.name),
                    "is_dir": member.isdir(),
                },
                "details":
                {
                    "size": member.size,
                    "type": int(self.type_map[member.type]),
                    "modified": member.mtime,
                },
                "access":
                {
                    "gid": member.gid,
                    "group": member.gname,
                    "permissions": Permissions(mode=member.mode).dump(),
                    "uid": member.uid,
                    "user": member.uname,
                },
                "tar": raw_tar_info,
            }
        return Info(raw_info)

    def setinfo(self, path, info):
        self.check()
        raise errors.ResourceReadOnly(path)

    def listdir(self, path):
        self.check()
        path = relpath(path)
        if path:
            try:
                member = self._tar.getmember(path)
            except KeyError:
                six.raise_from(errors.ResourceNotFound(path), None)
            else:
                if not member.isdir():
                    six.raise_from(errors.DirectoryExpected(path), None)

        return [
            basename(member.name)
            for member in self._tar
            if dirname(member.path) == path
        ]

    def makedir(self, path, permissions=None, recreate=False):
        self.check()
        raise errors.ResourceReadOnly(path)

    def openbin(self, path, mode="r", buffering=-1, **options):
        self.check()
        path = relpath(normpath(path))

        if 'w' in mode or '+' in mode or 'a' in mode:
            raise errors.ResourceReadOnly(path)

        try:
            member = self._tar.getmember(path)
        except KeyError:
            six.raise_from(errors.ResourceNotFound(path), None)

        if not member.isfile():
            raise errors.FileExpected(path)

        rw = RawWrapper(self._tar.extractfile(member))

        if six.PY2: # Patch nonexistent file.flush in Python2
            def _flush():
                pass
            rw.flush = _flush

        return rw

    def remove(self, path):
        self.check()
        raise errors.ResourceReadOnly(path)

    def removedir(self, path):
        self.check()
        raise errors.ResourceReadOnly(path)

    def close(self):
        super(ReadTarFS, self).close()
        self._tar.close()

    def isclosed(self):
        return self._tar.closed


if __name__ == "__main__":  # pragma: nocover
    from fs.tree import render
    from fs.opener import open_fs

    with TarFS('tests.tar') as tar_fs:
        print(tar_fs.listdir('/'))
        print(tar_fs.listdir('/tests/'))
        print(tar_fs.gettext('tests/ttt/settings.ini'))
        render(tar_fs)
        print(tar_fs)
        print(repr(tar_fs))

    with TarFS("TarFS.tar", write=True) as tar_fs:
        tar_fs.makedirs('foo/bar')
        tar_fs.settext('foo/bar/baz.txt', 'Hello, World')
        print(tar_fs)
        print(repr(tar_fs))



# coding: utf-8
"""Defines the OSFSOpener."""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from .base import Opener

class OSFSOpener(Opener):
    protocols = ['file', 'osfs']

    def open_fs(self, fs_url, parse_result, writeable, create, cwd):
        from ..osfs import OSFS
        from os.path import abspath, expanduser, normpath, join
        _path = abspath(join(cwd, expanduser(parse_result.resource)))
        path = normpath(_path)
        osfs = OSFS(path, create=create)
        return osfs



# coding: utf-8
"""
fs.opener.errors
================

Errors raised when attempting to open a filesystem.

"""

class ParseError(ValueError):
    """Raised when attempting to parse an invalid FS URL."""


class OpenerError(Exception):
    """Base class for opener related errors."""


class UnsupportedProtocol(OpenerError):
    """May be raised if no opener could be found for a given
    protocol."""


class EntryPointError(OpenerError):
    """Raised by the registry when an entry point cannot be loaded."""
